#include <sodium.h>
#include <lua.h>
#include <lauxlib.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
/* this file is automatically generated */



typedef struct luasodium_constant_s {
    const char *name;
    size_t value;
} luasodium_constant_t;

#define LS_CONST(x) { #x, x }

typedef void (*ls_func_ptr)(void);

/* base type for function definitions */
typedef struct luasodium_function_s {
    const char *name;
    ls_func_ptr func;
} luasodium_function_t;

#define LS_FUNC(x) { #x, (ls_func_ptr)x }

static void
luasodium_set_constants(lua_State *L, const luasodium_constant_t *c,int index) {
    for(; c->name != NULL; c++) {
        lua_pushinteger(L,c->value);
        lua_setfield(L,index,c->name);
    }
}



#define LS_LUA_FUNC(x) { #x, ls_ ## x }

#if !defined(luaL_newlibtable) \
  && (!defined LUA_VERSION_NUM || LUA_VERSION_NUM==501)
static void luaL_setfuncs (lua_State *L, const luaL_Reg *l, int nup) {
  luaL_checkstack(L, nup+1, "too many upvalues");
  for (; l->name != NULL; l++) {  /* fill the table with given functions */
    int i;
    lua_pushstring(L, l->name);
    for (i = 0; i < nup; i++)  /* copy upvalues to the top */
      lua_pushvalue(L, -(nup+1));
    lua_pushcclosure(L, l->func, nup);  /* closure with those upvalues */
    lua_settable(L, -(nup + 3));
  }
  lua_pop(L, nup);  /* remove upvalues */
}
#endif

#define LUASODIUM_INIT(L) \
if(sodium_init() == -1) { \
    lua_pushliteral(L,"sodium_init error"); \
    return lua_error(L); \
}

#if LUA_VERSION_NUM >= 502
static int
ls_lua_equal(lua_State *L, int index1, int index2) {
    return lua_compare(L,index1,index2,LUA_OPEQ);
}
#else
#define ls_lua_equal(L,index1,index2) lua_equal(L,index1,index2)
#endif


static const luasodium_constant_t ls_crypto_auth_constants[] = {
    LS_CONST(crypto_auth_BYTES),
    LS_CONST(crypto_auth_KEYBYTES),
    { NULL, 0 },
};



static int
ls_crypto_auth(lua_State *L) {
    unsigned char out[crypto_auth_BYTES];
    const unsigned char *in = NULL;
    const unsigned char *k = NULL;

    size_t inlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,2)) {
        lua_pushliteral(L,"requires 2 parameters");
        return lua_error(L);
    }

    in = (const unsigned char *)lua_tolstring(L,1,&inlen);
    k =  (const unsigned char *)lua_tolstring(L,2,&klen);

    if(klen != crypto_auth_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d", crypto_auth_KEYBYTES);
    }

    if(crypto_auth(out,in,inlen,k) == -1) {
        return luaL_error(L,"crypto_auth error");
    }
    lua_pushlstring(L,(const char *)out,crypto_auth_BYTES);
    sodium_memzero(out,crypto_auth_BYTES);
    return 1;
}

static int
ls_crypto_auth_verify(lua_State *L) {
    const unsigned char *h = NULL;
    const unsigned char *in = NULL;
    const unsigned char *k = NULL;

    size_t hlen = 0;
    size_t inlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 parameters");
        return lua_error(L);
    }

    h =  (const unsigned char *)lua_tolstring(L,1,&hlen);
    in = (const unsigned char *)lua_tolstring(L,2,&inlen);
    k =  (const unsigned char *)lua_tolstring(L,3,&klen);

    if(hlen != crypto_auth_BYTES) {
        return luaL_error(L,"wrong tag size, expected: %d", crypto_auth_BYTES);
    }

    if(klen != crypto_auth_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d", crypto_auth_KEYBYTES);
    }

    lua_pushboolean(L,crypto_auth_verify(h,in,inlen,k) == 0);
    return 1;
}

static int
ls_crypto_auth_keygen(lua_State *L) {
    unsigned char k[crypto_auth_KEYBYTES];
    crypto_auth_keygen(k);
    lua_pushlstring(L,(const char *)k,crypto_auth_KEYBYTES);
    sodium_memzero(k,crypto_auth_KEYBYTES);
    return 1;
}

static const struct luaL_Reg ls_crypto_auth_functions[] = {
    LS_LUA_FUNC(crypto_auth),
    LS_LUA_FUNC(crypto_auth_verify),
    LS_LUA_FUNC(crypto_auth_keygen),
    { NULL, NULL },
};

static int
ls_crypto_auth_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_crypto_auth_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_crypto_auth_functions,0);
    return 0;
}
static const luasodium_constant_t ls_crypto_box_constants[] = {
    LS_CONST(crypto_box_PUBLICKEYBYTES),
    LS_CONST(crypto_box_SECRETKEYBYTES),
    LS_CONST(crypto_box_MACBYTES),
    LS_CONST(crypto_box_NONCEBYTES),
    LS_CONST(crypto_box_SEEDBYTES),
    LS_CONST(crypto_box_BEFORENMBYTES),
    LS_CONST(crypto_box_BOXZEROBYTES),
    LS_CONST(crypto_box_ZEROBYTES),

    LS_CONST(crypto_box_curve25519xsalsa20poly1305_PUBLICKEYBYTES),
    LS_CONST(crypto_box_curve25519xsalsa20poly1305_SECRETKEYBYTES),
    LS_CONST(crypto_box_curve25519xsalsa20poly1305_MACBYTES),
    LS_CONST(crypto_box_curve25519xsalsa20poly1305_NONCEBYTES),
    LS_CONST(crypto_box_curve25519xsalsa20poly1305_SEEDBYTES),
    LS_CONST(crypto_box_curve25519xsalsa20poly1305_BEFORENMBYTES),
    LS_CONST(crypto_box_curve25519xsalsa20poly1305_BOXZEROBYTES),
    LS_CONST(crypto_box_curve25519xsalsa20poly1305_ZEROBYTES),

    LS_CONST(crypto_box_curve25519xchacha20poly1305_PUBLICKEYBYTES),
    LS_CONST(crypto_box_curve25519xchacha20poly1305_SECRETKEYBYTES),
    LS_CONST(crypto_box_curve25519xchacha20poly1305_MACBYTES),
    LS_CONST(crypto_box_curve25519xchacha20poly1305_NONCEBYTES),
    LS_CONST(crypto_box_curve25519xchacha20poly1305_SEEDBYTES),
    LS_CONST(crypto_box_curve25519xchacha20poly1305_BEFORENMBYTES),
    { NULL, 0 },
};



static int
ls_crypto_box_keypair(lua_State *L) {
    unsigned char pk[crypto_box_PUBLICKEYBYTES];
    unsigned char sk[crypto_box_SECRETKEYBYTES];

    if(crypto_box_keypair(pk,sk) == -1) {
        lua_pushliteral(L,"crypto_box_keypair error");
        return lua_error(L);
    }

    lua_pushlstring(L,(const char *)pk,crypto_box_PUBLICKEYBYTES);
    lua_pushlstring(L,(const char *)sk,crypto_box_SECRETKEYBYTES);

    sodium_memzero(pk,crypto_box_PUBLICKEYBYTES);
    sodium_memzero(sk,crypto_box_SECRETKEYBYTES);

    return 2;
}

static int
ls_crypto_box_seed_keypair(lua_State *L) {
    unsigned char pk[crypto_box_PUBLICKEYBYTES];
    unsigned char sk[crypto_box_SECRETKEYBYTES];
    const unsigned char *seed = NULL;
    size_t seed_len = 0;

    if(lua_isnoneornil(L,1)) {
        lua_pushliteral(L,"requires 1 argument");
        return lua_error(L);
    }

    seed = (const unsigned char *)lua_tolstring(L,1,&seed_len);

    if(seed_len != crypto_box_SEEDBYTES) {
        return luaL_error(L,"wrong seed length, expected: %d",
          crypto_box_SEEDBYTES);
    }

    if(crypto_box_seed_keypair(pk,sk,seed) == -1) {
        return luaL_error(L,"crypto_box_seed_keypair error");
    }
    lua_pushlstring(L,(const char *)pk,crypto_box_PUBLICKEYBYTES);
    lua_pushlstring(L,(const char *)sk,crypto_box_SECRETKEYBYTES);

    sodium_memzero(pk,crypto_box_PUBLICKEYBYTES);
    sodium_memzero(sk,crypto_box_SECRETKEYBYTES);

    return 2;
}

static int
ls_crypto_box(lua_State *L) {
    unsigned char *c      = NULL;
    const unsigned char *m = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *pk    = NULL;
    const unsigned char *sk    = NULL;

    unsigned char *tmp_m   = NULL;

    size_t clen = 0;
    size_t mlen = 0;
    size_t noncelen = 0;
    size_t pklen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,4)) {
        lua_pushliteral(L,"requires 4 parameters");
        return lua_error(L);
    }

    m = (const unsigned char *)lua_tolstring(L,1,&mlen);
    nonce = (const unsigned char *)lua_tolstring(L,2,&noncelen);
    pk    = (const unsigned char *)lua_tolstring(L,3,&pklen);
    sk    = (const unsigned char *)lua_tolstring(L,4,&sklen);

    if(noncelen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_box_NONCEBYTES);
    }

    if(pklen != crypto_box_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key size, expected: %d",crypto_box_PUBLICKEYBYTES);
    }

    if(sklen != crypto_box_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key size, expected: %d",crypto_box_SECRETKEYBYTES);
    }

    clen = mlen + crypto_box_MACBYTES;

    tmp_m = (unsigned char *)lua_newuserdata(L,mlen + crypto_box_ZEROBYTES);
    if(tmp_m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    c = (unsigned char *)lua_newuserdata(L,clen + crypto_box_BOXZEROBYTES);
    if(c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,2);

    sodium_memzero(tmp_m,crypto_box_ZEROBYTES);
    sodium_memzero(c,crypto_box_BOXZEROBYTES);

    memcpy(&tmp_m[crypto_box_ZEROBYTES],m,mlen);

    if(crypto_box(c,tmp_m,mlen+crypto_box_ZEROBYTES,nonce,pk,sk) == -1) {
        return luaL_error(L,"crypto_box error");
    }

    lua_pushlstring(L,(const char *)&c[crypto_box_BOXZEROBYTES],clen);
    sodium_memzero(tmp_m,mlen + crypto_box_ZEROBYTES);
    sodium_memzero(c,clen + crypto_box_BOXZEROBYTES);
    return 1;
}

static int
ls_crypto_box_open(lua_State *L) {
    unsigned char *m      = NULL;
    const unsigned char *c = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *pk    = NULL;
    const unsigned char *sk    = NULL;

    unsigned char *tmp_c   = NULL;

    size_t mlen = 0;
    size_t clen = 0;
    size_t noncelen = 0;
    size_t pklen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,4)) {
        lua_pushliteral(L,"requires 4 parameters");
        return lua_error(L);
    }

    c = (const unsigned char *)lua_tolstring(L,1,&clen);
    nonce = (const unsigned char *)lua_tolstring(L,2,&noncelen);
    pk    = (const unsigned char *)lua_tolstring(L,3,&pklen);
    sk    = (const unsigned char *)lua_tolstring(L,4,&sklen);

    if(clen <= crypto_box_MACBYTES) {
        return luaL_error(L,"wrong mac size, expected at least: %d",crypto_box_MACBYTES);
    }

    if(noncelen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_box_NONCEBYTES);
    }

    if(pklen != crypto_box_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key size, expected: %d",crypto_box_PUBLICKEYBYTES);
    }

    if(sklen != crypto_box_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key size, expected: %d",crypto_box_SECRETKEYBYTES);
    }

    mlen = clen - crypto_box_MACBYTES;

    tmp_c = (unsigned char *)lua_newuserdata(L,clen + crypto_box_BOXZEROBYTES);
    if(tmp_c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    m = (unsigned char *)lua_newuserdata(L,mlen + crypto_box_ZEROBYTES);
    if(m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,2);

    sodium_memzero(tmp_c,crypto_box_BOXZEROBYTES);
    sodium_memzero(m,crypto_box_ZEROBYTES);

    memcpy(&tmp_c[crypto_box_BOXZEROBYTES],c,clen);

    if(crypto_box_open(m,tmp_c,clen+crypto_box_BOXZEROBYTES,nonce,pk,sk) == -1) {
        return luaL_error(L,"crypto_box_open error");
    }

    lua_pushlstring(L,(const char *)&m[crypto_box_ZEROBYTES],mlen);
    sodium_memzero(tmp_c,clen + crypto_box_BOXZEROBYTES);
    sodium_memzero(m,mlen + crypto_box_ZEROBYTES);
    return 1;
}

static int
ls_crypto_box_easy(lua_State *L) {
    unsigned char *c = NULL;
    const unsigned char *m = NULL;
    const unsigned char *n = NULL;
    const unsigned char *pk = NULL;
    const unsigned char *sk = NULL;
    size_t clen = 0;
    size_t mlen = 0;
    size_t nlen = 0;
    size_t pklen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,4)) {
        lua_pushliteral(L,"requires 4 arguments");
        return lua_error(L);
    }

    m  = (const unsigned char *)lua_tolstring(L,1,&mlen);
    n  = (const unsigned char *)lua_tolstring(L,2,&nlen);
    pk = (const unsigned char *)lua_tolstring(L,3,&pklen);
    sk = (const unsigned char *)lua_tolstring(L,4,&sklen);

    if(nlen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce length, expected: %d",
          crypto_box_NONCEBYTES);
    }

    if(pklen != crypto_box_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key length, expected: %d",
          crypto_box_PUBLICKEYBYTES);
    }

    if(sklen != crypto_box_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key length, expected: %d",
          crypto_box_SECRETKEYBYTES);
    }

    clen = mlen + crypto_box_MACBYTES;

    c = lua_newuserdata(L,clen);
    if(c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_box_easy(c,m,mlen,n,pk,sk) == -1) {
        return luaL_error(L,"crypto_box_easy error");
    }

    lua_pushlstring(L,(const char *)c,clen);
    sodium_memzero(c,clen);
    return 1;
}

static int
ls_crypto_box_open_easy(lua_State *L) {
    unsigned char *m = NULL;
    const unsigned char *c = NULL;
    const unsigned char *n = NULL;
    const unsigned char *pk = NULL;
    const unsigned char *sk = NULL;
    size_t mlen = 0;
    size_t clen = 0;
    size_t nlen = 0;
    size_t pklen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,4)) {
        lua_pushliteral(L,"requires 4 arguments");
        return lua_error(L);
    }

    c  = (const unsigned char *)lua_tolstring(L,1,&clen);
    n  = (const unsigned char *)lua_tolstring(L,2,&nlen);
    pk = (const unsigned char *)lua_tolstring(L,3,&pklen);
    sk = (const unsigned char *)lua_tolstring(L,4,&sklen);

    if(clen <= crypto_box_MACBYTES) {
        return luaL_error(L,"wrong cipher length, expected at least: %d",
          crypto_box_MACBYTES);
    }

    if(nlen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce length, expected: %d",
          crypto_box_NONCEBYTES);
    }

    if(pklen != crypto_box_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key length, expected: %d",
          crypto_box_PUBLICKEYBYTES);
    }

    if(sklen != crypto_box_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key length, expected: %d",
          crypto_box_SECRETKEYBYTES);
    }

    mlen = clen - crypto_box_MACBYTES;

    m = lua_newuserdata(L,mlen);
    if(m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_box_open_easy(m,c,clen,n,pk,sk) == -1) {
        return luaL_error(L,"crypto_box_open_easy error");
    }

    lua_pushlstring(L,(const char *)m,mlen);
    sodium_memzero(m,mlen);
    return 1;
}

static int
ls_crypto_box_detached(lua_State *L) {
    unsigned char *c   = NULL;
    unsigned char mac[crypto_box_MACBYTES];
    const unsigned char *m = NULL;
    const unsigned char *n = NULL;
    const unsigned char *pk = NULL;
    const unsigned char *sk = NULL;
    size_t mlen = 0;
    size_t nlen = 0;
    size_t pklen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,4)) {
        lua_pushliteral(L,"requires 4 arguments");
        return lua_error(L);
    }

    m  = (const unsigned char *)lua_tolstring(L,1,&mlen);
    n  = (const unsigned char *)lua_tolstring(L,2,&nlen);
    pk = (const unsigned char *)lua_tolstring(L,3,&pklen);
    sk = (const unsigned char *)lua_tolstring(L,4,&sklen);

    if(nlen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce length, expected: %d",
          crypto_box_NONCEBYTES);
    }

    if(pklen != crypto_box_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key length, expected: %d",
          crypto_box_PUBLICKEYBYTES);
    }

    if(sklen != crypto_box_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key length, expected: %d",
          crypto_box_SECRETKEYBYTES);
    }

    c = lua_newuserdata(L,mlen);
    if(c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_box_detached(c,mac,m,mlen,n,pk,sk) == -1) {
        return luaL_error(L,"crypto_box_detached error");
    }

    lua_pushlstring(L,(const char *)c,mlen);
    lua_pushlstring(L,(const char *)mac,crypto_box_MACBYTES);

    sodium_memzero(c,mlen);
    sodium_memzero(mac,crypto_box_MACBYTES);
    return 2;
}

static int
ls_crypto_box_open_detached(lua_State *L) {
    unsigned char *m = NULL;
    const unsigned char *c = NULL;
    const unsigned char *mac = NULL;
    const unsigned char *n = NULL;
    const unsigned char *pk = NULL;
    const unsigned char *sk = NULL;
    size_t clen = 0;
    size_t maclen = 0;
    size_t nlen = 0;
    size_t pklen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,5)) {
        lua_pushliteral(L,"requires 5 arguments");
        return lua_error(L);
    }

    c   = (const unsigned char *)lua_tolstring(L,1,&clen);
    mac = (const unsigned char *)lua_tolstring(L,2,&maclen);
    n   = (const unsigned char *)lua_tolstring(L,3,&nlen);
    pk  = (const unsigned char *)lua_tolstring(L,4,&pklen);
    sk  = (const unsigned char *)lua_tolstring(L,5,&sklen);

    if(maclen != crypto_box_MACBYTES) {
        return luaL_error(L,"wrong mac length, expected: %d",
          crypto_box_MACBYTES);
    }

    if(nlen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce length, expected: %d",
          crypto_box_NONCEBYTES);
    }

    if(pklen != crypto_box_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key length, expected: %d",
          crypto_box_PUBLICKEYBYTES);
    }

    if(sklen != crypto_box_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key length, expected: %d",
          crypto_box_SECRETKEYBYTES);
    }

    m = lua_newuserdata(L,clen);
    if(m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_box_open_detached(m,c,mac,clen,n,pk,sk) == -1) {
        return luaL_error(L,"crypto_box_open_detached error");
    }

    lua_pushlstring(L,(const char *)m,clen);

    sodium_memzero(m,clen);

    return 1;
}

static int
ls_crypto_box_beforenm(lua_State *L) {
    unsigned char k[crypto_box_BEFORENMBYTES];
    const unsigned char *pk = NULL;
    const unsigned char *sk = NULL;

    size_t pklen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,2)) {
        lua_pushliteral(L,"requires 2 arguments");
        return lua_error(L);
    }

    pk  = (const unsigned char *)lua_tolstring(L,1,&pklen);
    sk  = (const unsigned char *)lua_tolstring(L,2,&sklen);

    if(pklen != crypto_box_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key length, expected: %d",
          crypto_box_PUBLICKEYBYTES);
    }

    if(sklen != crypto_box_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key length, expected: %d",
          crypto_box_SECRETKEYBYTES);
    }

    if(crypto_box_beforenm(k,pk,sk) == -1) {
        return luaL_error(L,"crypto_box_beforenm error");
    }

    lua_pushlstring(L,(const char *)k,crypto_box_BEFORENMBYTES);
    sodium_memzero(k,crypto_box_BEFORENMBYTES);
    return 1;
}

static int
ls_crypto_box_easy_afternm(lua_State *L) {
    unsigned char *c = NULL;
    const unsigned char *m = NULL;
    const unsigned char *n = NULL;
    const unsigned char *k = NULL;
    size_t clen = 0;
    size_t mlen = 0;
    size_t nlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 arguments");
        return lua_error(L);
    }

    m  = (const unsigned char *)lua_tolstring(L,1,&mlen);
    n  = (const unsigned char *)lua_tolstring(L,2,&nlen);
    k  = (const unsigned char *)lua_tolstring(L,3,&klen);

    if(nlen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce length, expected: %d",
          crypto_box_NONCEBYTES);
    }

    if(klen != crypto_box_BEFORENMBYTES) {
        return luaL_error(L,"wrong shared key length, expected: %d",
          crypto_box_BEFORENMBYTES);
    }

    clen = mlen + crypto_box_MACBYTES;

    c = lua_newuserdata(L,clen);
    if(c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_box_easy_afternm(c,m,mlen,n,k) == -1) {
        return luaL_error(L,"crypto_box_easy_afternm error");
    }

    lua_pushlstring(L,(const char *)c,clen);
    sodium_memzero(c,clen);
    return 1;
}

static int
ls_crypto_box_open_easy_afternm(lua_State *L) {
    unsigned char *m = NULL;
    const unsigned char *c = NULL;
    const unsigned char *n = NULL;
    const unsigned char *k = NULL;
    size_t mlen = 0;
    size_t clen = 0;
    size_t nlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 arguments");
        return lua_error(L);
    }

    c  = (const unsigned char *)lua_tolstring(L,1,&clen);
    n  = (const unsigned char *)lua_tolstring(L,2,&nlen);
    k  = (const unsigned char *)lua_tolstring(L,3,&klen);

    if(clen <= crypto_box_MACBYTES) {
        return luaL_error(L,"wrong cipher length, expected at least: %d",
          crypto_box_MACBYTES);
    }

    if(nlen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce length, expected: %d",
          crypto_box_NONCEBYTES);
    }

    if(klen != crypto_box_BEFORENMBYTES) {
        return luaL_error(L,"wrong shared key length, expected: %d",
          crypto_box_BEFORENMBYTES);
    }

    mlen = clen - crypto_box_MACBYTES;

    m = lua_newuserdata(L,mlen);
    if(m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_box_open_easy_afternm(m,c,clen,n,k) == -1) {
        return luaL_error(L,"crypto_box_open_easy_afternm error");
    }

    lua_pushlstring(L,(const char *)m,mlen);
    sodium_memzero(m,mlen);
    return 1;
}

static int
ls_crypto_box_detached_afternm(lua_State *L) {
    unsigned char *c = NULL;
    unsigned char mac[crypto_box_MACBYTES];
    const unsigned char *m = NULL;
    const unsigned char *n = NULL;
    const unsigned char *k = NULL;
    size_t mlen = 0;
    size_t nlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 arguments");
        return lua_error(L);
    }

    m  = (const unsigned char *)lua_tolstring(L,1,&mlen);
    n  = (const unsigned char *)lua_tolstring(L,2,&nlen);
    k  = (const unsigned char *)lua_tolstring(L,3,&klen);

    if(nlen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce length, expected: %d",
          crypto_box_NONCEBYTES);
    }

    if(klen != crypto_box_BEFORENMBYTES) {
        return luaL_error(L,"wrong shared key length, expected: %d",
          crypto_box_BEFORENMBYTES);
    }

    c = lua_newuserdata(L,mlen);
    if(c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_box_detached_afternm(c,mac,m,mlen,n,k) == -1) {
        return luaL_error(L,"crypto_box_detached_afternm error");
    }

    lua_pushlstring(L,(const char *)c,mlen);
    lua_pushlstring(L,(const char *)mac,crypto_box_MACBYTES);

    sodium_memzero(c,mlen);
    sodium_memzero(mac,crypto_box_MACBYTES);
    return 2;
}

static int
ls_crypto_box_open_detached_afternm(lua_State *L) {
    unsigned char *m = NULL;
    const unsigned char *c   = NULL;
    const unsigned char *mac = NULL;
    const unsigned char *n   = NULL;
    const unsigned char *k   = NULL;
    size_t clen = 0;
    size_t maclen = 0;
    size_t nlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,4)) {
        lua_pushliteral(L,"requires 4 arguments");
        return lua_error(L);
    }

    c   = (const unsigned char *)lua_tolstring(L,1,&clen);
    mac = (const unsigned char *)lua_tolstring(L,2,&maclen);
    n   = (const unsigned char *)lua_tolstring(L,3,&nlen);
    k   = (const unsigned char *)lua_tolstring(L,4,&klen);

    if(maclen != crypto_box_MACBYTES) {
        return luaL_error(L,"wrong mac length, expected: %d",
          crypto_box_MACBYTES);
    }

    if(nlen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce length, expected: %d",
          crypto_box_NONCEBYTES);
    }

    if(klen != crypto_box_BEFORENMBYTES) {
        return luaL_error(L,"wrong shared key length, expected: %d",
          crypto_box_BEFORENMBYTES);
    }

    m = lua_newuserdata(L,clen);
    if(m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_box_open_detached_afternm(m,c,mac,clen,n,k) == -1) {
        return luaL_error(L,"crypto_box_open_detached error");
    }

    lua_pushlstring(L,(const char *)m,clen);
    return 1;
}

static int
ls_crypto_box_afternm(lua_State *L) {
    unsigned char *c      = NULL;
    const unsigned char *m = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *k    = NULL;

    unsigned char *tmp_m   = NULL;

    size_t clen = 0;
    size_t mlen = 0;
    size_t noncelen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 parameters");
        return lua_error(L);
    }

    m     = (const unsigned char *)lua_tolstring(L,1,&mlen);
    nonce = (const unsigned char *)lua_tolstring(L,2,&noncelen);
    k     = (const unsigned char *)lua_tolstring(L,3,&klen);

    if(noncelen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_box_NONCEBYTES);
    }

    if(klen != crypto_box_BEFORENMBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",crypto_box_BEFORENMBYTES);
    }

    clen = mlen + crypto_box_MACBYTES;

    tmp_m = (unsigned char *)lua_newuserdata(L,mlen + crypto_box_ZEROBYTES);
    if(tmp_m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    c = (unsigned char *)lua_newuserdata(L,clen + crypto_box_BOXZEROBYTES);
    if(c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,2);

    sodium_memzero(tmp_m,crypto_box_ZEROBYTES);
    sodium_memzero(c,crypto_box_BOXZEROBYTES);

    memcpy(&tmp_m[crypto_box_ZEROBYTES],m,mlen);

    if(crypto_box_afternm(c,tmp_m,mlen+crypto_box_ZEROBYTES,nonce,k) == -1) {
        return luaL_error(L,"crypto_box_afternm error");
    }

    lua_pushlstring(L,(const char *)&c[crypto_box_BOXZEROBYTES],clen);
    sodium_memzero(tmp_m,mlen + crypto_box_ZEROBYTES);
    sodium_memzero(c,clen + crypto_box_BOXZEROBYTES);
    return 1;
}

static int
ls_crypto_box_open_afternm(lua_State *L) {
    unsigned char *m      = NULL;
    const unsigned char *c = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *k    = NULL;

    unsigned char *tmp_c   = NULL;

    size_t mlen = 0;
    size_t clen = 0;
    size_t noncelen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 parameters");
        return lua_error(L);
    }

    c = (const unsigned char *)lua_tolstring(L,1,&clen);
    nonce = (const unsigned char *)lua_tolstring(L,2,&noncelen);
    k     = (const unsigned char *)lua_tolstring(L,3,&klen);

    if(clen <= crypto_box_MACBYTES) {
        return luaL_error(L,"wrong mac size, expected at least: %d",crypto_box_MACBYTES);
    }

    if(noncelen != crypto_box_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_box_NONCEBYTES);
    }

    if(klen != crypto_box_BEFORENMBYTES) {
        return luaL_error(L,"wrong public key size, expected: %d",crypto_box_BEFORENMBYTES);
    }

    mlen = clen - crypto_box_MACBYTES;

    tmp_c = (unsigned char *)lua_newuserdata(L,clen + crypto_box_BOXZEROBYTES);
    if(tmp_c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    m = (unsigned char *)lua_newuserdata(L,mlen + crypto_box_ZEROBYTES);
    if(m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,2);

    sodium_memzero(tmp_c,crypto_box_BOXZEROBYTES);
    sodium_memzero(m,crypto_box_ZEROBYTES);

    memcpy(&tmp_c[crypto_box_BOXZEROBYTES],c,clen);

    if(crypto_box_open_afternm(m,tmp_c,clen+crypto_box_BOXZEROBYTES,nonce,k) == -1) {
        return luaL_error(L,"crypto_box_open_afternm error");
    }

    lua_pushlstring(L,(const char *)&m[crypto_box_ZEROBYTES],mlen);
    sodium_memzero(tmp_c,clen + crypto_box_BOXZEROBYTES);
    sodium_memzero(m,mlen + crypto_box_ZEROBYTES);
    return 1;
}

static const struct luaL_Reg ls_crypto_box_functions[] = {
    LS_LUA_FUNC(crypto_box),
    LS_LUA_FUNC(crypto_box_open),
    LS_LUA_FUNC(crypto_box_easy),
    LS_LUA_FUNC(crypto_box_open_easy),
    LS_LUA_FUNC(crypto_box_detached),
    LS_LUA_FUNC(crypto_box_open_detached),
    LS_LUA_FUNC(crypto_box_beforenm),
    LS_LUA_FUNC(crypto_box_easy_afternm),
    LS_LUA_FUNC(crypto_box_open_easy_afternm),
    LS_LUA_FUNC(crypto_box_detached_afternm),
    LS_LUA_FUNC(crypto_box_open_detached_afternm),
    LS_LUA_FUNC(crypto_box_keypair),
    LS_LUA_FUNC(crypto_box_seed_keypair),
    LS_LUA_FUNC(crypto_box_afternm),
    LS_LUA_FUNC(crypto_box_open_afternm),
    { NULL, NULL },
};

static int
ls_crypto_box_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_crypto_box_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_crypto_box_functions,0);
    return 0;
}

static const luasodium_constant_t ls_crypto_hash_constants[] = {
    LS_CONST(crypto_hash_BYTES),
    LS_CONST(crypto_hash_sha256_BYTES),
    LS_CONST(crypto_hash_sha512_BYTES),
    { NULL, 0 },
};




static int
ls_crypto_hash(lua_State *L) {
    unsigned char h[crypto_hash_BYTES];
    const unsigned char *m = NULL;
    size_t mlen = 0;

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameter");
    }

    m = (const unsigned char *)lua_tolstring(L,1,&mlen);

    if(crypto_hash(h,m,mlen) == -1) {
        return luaL_error(L,"crypto_hash error");
    }

    lua_pushlstring(L,(const char *)h,crypto_hash_BYTES);
    sodium_memzero(h,crypto_hash_BYTES);
    return 1;
}

static int
ls_crypto_hash_sha256(lua_State *L) {
    unsigned char h[crypto_hash_sha256_BYTES];
    const unsigned char *m = NULL;
    size_t mlen = 0;

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameter");
    }

    m = (const unsigned char *)lua_tolstring(L,1,&mlen);

    if(crypto_hash_sha256(h,m,mlen) == -1) {
        return luaL_error(L,"crypto_hash error");
    }

    lua_pushlstring(L,(const char *)h,crypto_hash_sha256_BYTES);
    sodium_memzero(h,crypto_hash_sha256_BYTES);
    return 1;
}

static int
ls_crypto_hash_sha512(lua_State *L) {
    unsigned char h[crypto_hash_sha512_BYTES];
    const unsigned char *m = NULL;
    size_t mlen = 0;

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameter");
    }

    m = (const unsigned char *)lua_tolstring(L,1,&mlen);

    if(crypto_hash_sha512(h,m,mlen) == -1) {
        return luaL_error(L,"crypto_hash error");
    }

    lua_pushlstring(L,(const char *)h,crypto_hash_sha512_BYTES);
    sodium_memzero(h,crypto_hash_sha512_BYTES);
    return 1;
}

static int
ls_crypto_hash_sha256_init(lua_State *L) {
    crypto_hash_sha256_state *state = NULL;
    state = (crypto_hash_sha256_state *)lua_newuserdata(L,
      crypto_hash_sha256_statebytes());

    if(state == NULL) {
        return luaL_error(L,"out of memory");
    }

    if(crypto_hash_sha256_init(state) == -1) {
        lua_pop(L,1);
        return luaL_error(L,"crypto_hash_sha256_init error");
    }

    lua_pushvalue(L,lua_upvalueindex(1));
    lua_setmetatable(L,-2);

    return 1;
}

static int
ls_crypto_hash_sha256_update(lua_State *L) {
    crypto_hash_sha256_state *state = NULL;
    const unsigned char *m   = NULL;
    size_t mlen = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 parameters");
    }

    /* verify this is a state object */
    lua_pushvalue(L,lua_upvalueindex(1));
    lua_getmetatable(L,1);

    if(!ls_lua_equal(L,-2,-1)) {
        return luaL_error(L,"invalid userdata");
    }
    lua_pop(L,2);

    state = (crypto_hash_sha256_state *)lua_touserdata(L,1);
    m   = (const unsigned char *)lua_tolstring(L,2,&mlen);

    lua_pushboolean(L,crypto_hash_sha256_update(
      state,m,mlen) != -1);
    return 1;
}

static int
ls_crypto_hash_sha256_final(lua_State *L) {
    crypto_hash_sha256_state *state = NULL;
    unsigned char h[crypto_hash_sha256_BYTES];

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameters");
    }

    /* verify this is a state object */
    lua_pushvalue(L,lua_upvalueindex(1));
    lua_getmetatable(L,1);

    if(!ls_lua_equal(L,-2,-1)) {
        return luaL_error(L,"invalid userdata");
    }
    lua_pop(L,2);

    state = (crypto_hash_sha256_state *)lua_touserdata(L,1);

    if(crypto_hash_sha256_final(state,h) == -1) {
        return luaL_error(L,"crypto_hash_sha256_final error");
    }

    lua_pushlstring(L,(const char *)h,crypto_hash_sha256_BYTES);
    sodium_memzero(h,crypto_hash_sha256_BYTES);

    return 1;
}

static int
ls_crypto_hash_sha512_init(lua_State *L) {
    crypto_hash_sha512_state *state = NULL;
    state = (crypto_hash_sha512_state *)lua_newuserdata(L,
      crypto_hash_sha512_statebytes());

    if(state == NULL) {
        return luaL_error(L,"out of memory");
    }

    if(crypto_hash_sha512_init(state) == -1) {
        lua_pop(L,1);
        return luaL_error(L,"crypto_hash_sha512_init error");
    }

    lua_pushvalue(L,lua_upvalueindex(1));
    lua_setmetatable(L,-2);

    return 1;
}

static int
ls_crypto_hash_sha512_update(lua_State *L) {
    crypto_hash_sha512_state *state = NULL;
    const unsigned char *m   = NULL;
    size_t mlen = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 parameters");
    }

    /* verify this is a state object */
    lua_pushvalue(L,lua_upvalueindex(1));
    lua_getmetatable(L,1);

    if(!ls_lua_equal(L,-2,-1)) {
        return luaL_error(L,"invalid userdata");
    }
    lua_pop(L,2);

    state = (crypto_hash_sha512_state *)lua_touserdata(L,1);
    m   = (const unsigned char *)lua_tolstring(L,2,&mlen);

    lua_pushboolean(L,crypto_hash_sha512_update(
      state,m,mlen) != -1);
    return 1;
}

static int
ls_crypto_hash_sha512_final(lua_State *L) {
    crypto_hash_sha512_state *state = NULL;
    unsigned char h[crypto_hash_sha512_BYTES];

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameters");
    }

    /* verify this is a state object */
    lua_pushvalue(L,lua_upvalueindex(1));
    lua_getmetatable(L,1);

    if(!ls_lua_equal(L,-2,-1)) {
        return luaL_error(L,"invalid userdata");
    }
    lua_pop(L,2);

    state = (crypto_hash_sha512_state *)lua_touserdata(L,1);

    if(crypto_hash_sha512_final(state,h) == -1) {
        return luaL_error(L,"crypto_hash_sha512_final error");
    }

    lua_pushlstring(L,(const char *)h,crypto_hash_sha512_BYTES);
    sodium_memzero(h,crypto_hash_sha512_BYTES);

    return 1;
}

static int
ls_crypto_hash_sha256_state__gc(lua_State *L) {
    crypto_hash_sha256_state *state = (crypto_hash_sha256_state *)lua_touserdata(L,1);
    sodium_memzero(state,crypto_hash_sha256_statebytes());
    return 0;
}

static int
ls_crypto_hash_sha512_state__gc(lua_State *L) {
    crypto_hash_sha512_state *state = (crypto_hash_sha512_state *)lua_touserdata(L,1);
    sodium_memzero(state,crypto_hash_sha512_statebytes());
    return 0;
}

static const struct luaL_Reg ls_crypto_hash_functions[] = {
    LS_LUA_FUNC(crypto_hash),
    LS_LUA_FUNC(crypto_hash_sha256),
    LS_LUA_FUNC(crypto_hash_sha512),
    LS_LUA_FUNC(crypto_hash_sha256_init),
    { NULL, NULL },
};


static const struct luaL_Reg ls_crypto_hash_sha256_state_functions[] = {
    LS_LUA_FUNC(crypto_hash_sha256_init),
    LS_LUA_FUNC(crypto_hash_sha256_update),
    LS_LUA_FUNC(crypto_hash_sha256_final),
    { NULL, NULL },
};

static const struct luaL_Reg ls_crypto_hash_sha512_state_functions[] = {
    LS_LUA_FUNC(crypto_hash_sha512_init),
    LS_LUA_FUNC(crypto_hash_sha512_update),
    LS_LUA_FUNC(crypto_hash_sha512_final),
    { NULL, NULL },
};

static int
ls_crypto_hash_sha256_state_setup(lua_State *L) {

    /* create our metatable for crypto_hash_sha256_state */
    lua_newtable(L);
    lua_pushcclosure(L,ls_crypto_hash_sha256_state__gc,0);
    lua_setfield(L,-2,"__gc");

    /* table of methods */
    lua_newtable(L);
    lua_setfield(L,-2,"__index");

    /* top of stack is our metatable */
    /* push up copies of our module + metatable since setfuncs will pop metatable */
    lua_pushvalue(L,-2); /* module */
    lua_pushvalue(L,-2); /* metatable */
    luaL_setfuncs(L,ls_crypto_hash_sha256_state_functions,1);
    lua_pop(L,1); /* module (copy) */

    /* stack is now:
     *   table (our modules)
     *   table (our metatable)
     */

    lua_getfield(L,-1,"__index");
    /* module
     * metatable
     * __index
     */

    lua_getfield(L,-3,"crypto_hash_sha256_update");
    /* module
     * metatable
     * __index
     * function
     */
    lua_setfield(L,-2,"update");

    lua_getfield(L,-3,"crypto_hash_sha256_final");
    lua_setfield(L,-2,"final");

    /* module
     * metatable
     * __index
     */

    lua_pop(L,2);
    return 0;
}

static int
ls_crypto_hash_sha512_state_setup(lua_State *L) {

    /* create our metatable for crypto_hash_sha512_state */
    lua_newtable(L);
    lua_pushcclosure(L,ls_crypto_hash_sha512_state__gc,0);
    lua_setfield(L,-2,"__gc");

    /* table of methods */
    lua_newtable(L);
    lua_setfield(L,-2,"__index");

    /* top of stack is our metatable */
    /* push up copies of our module + metatable since setfuncs will pop metatable */
    lua_pushvalue(L,-2); /* module */
    lua_pushvalue(L,-2); /* metatable */
    luaL_setfuncs(L,ls_crypto_hash_sha512_state_functions,1);
    lua_pop(L,1); /* module (copy) */

    /* stack is now:
     *   table (our modules)
     *   table (our metatable)
     */

    lua_getfield(L,-1,"__index");
    /* module
     * metatable
     * __index
     */

    lua_getfield(L,-3,"crypto_hash_sha512_update");
    /* module
     * metatable
     * __index
     * function
     */
    lua_setfield(L,-2,"update");

    lua_getfield(L,-3,"crypto_hash_sha512_final");
    lua_setfield(L,-2,"final");

    /* module
     * metatable
     * __index
     */

    lua_pop(L,2);
    return 0;
}

static int
ls_crypto_hash_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_crypto_hash_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_crypto_hash_functions,0);
    ls_crypto_hash_sha256_state_setup(L);
    ls_crypto_hash_sha512_state_setup(L);
    return 0;
}

static const luasodium_constant_t ls_crypto_onetimeauth_constants[] = {
    LS_CONST(crypto_onetimeauth_BYTES),
    LS_CONST(crypto_onetimeauth_KEYBYTES),
    { NULL, 0 },
};


static int
ls_crypto_onetimeauth(lua_State *L) {
    unsigned char c[crypto_onetimeauth_BYTES];
    const unsigned char *m = NULL;
    const unsigned char *k = NULL;
    size_t mlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 arguments");
    }

    m = (const unsigned char *)lua_tolstring(L,1,&mlen);
    k = (const unsigned char *)lua_tolstring(L,2,&klen);

    if(klen != crypto_onetimeauth_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",
          crypto_onetimeauth_KEYBYTES);
    }

    if(crypto_onetimeauth(c,m,mlen,k) == -1) {
        return luaL_error(L,"crypto_onetimeauth error");
    }

    lua_pushlstring(L,(const char *)c,crypto_onetimeauth_BYTES);
    sodium_memzero(c,crypto_onetimeauth_BYTES);
    return 1;
}

static int
ls_crypto_onetimeauth_verify(lua_State *L) {
    const unsigned char *c = NULL;
    const unsigned char *m = NULL;
    const unsigned char *k = NULL;
    size_t clen = 0;
    size_t mlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,3)) {
        return luaL_error(L,"requires 3 arguments");
    }

    c = (const unsigned char *)lua_tolstring(L,1,&clen);
    m = (const unsigned char *)lua_tolstring(L,2,&mlen);
    k = (const unsigned char *)lua_tolstring(L,3,&klen);

    if(clen != crypto_onetimeauth_BYTES) {
        return luaL_error(L,"wrong authenticator size, expected: %d",
          crypto_onetimeauth_BYTES);
    }

    if(klen != crypto_onetimeauth_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",
          crypto_onetimeauth_KEYBYTES);
    }

    lua_pushboolean(L,crypto_onetimeauth_verify(c,m,mlen,k) == 0);
    return 1;
}

static int
ls_crypto_onetimeauth_keygen(lua_State *L) {
    unsigned char k[crypto_onetimeauth_KEYBYTES];

    crypto_onetimeauth_keygen(k);

    lua_pushlstring(L,(const char *)k,crypto_onetimeauth_KEYBYTES);
    sodium_memzero(k,crypto_onetimeauth_KEYBYTES);
    return 1;
}

static int
ls_crypto_onetimeauth_init(lua_State *L) {
    crypto_onetimeauth_state *state = NULL;
    const unsigned char *k = NULL;
    size_t klen = 0;

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameter");
    }

    k = (const unsigned char *)lua_tolstring(L,1,&klen);

    if(klen != crypto_onetimeauth_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %s",
          crypto_onetimeauth_KEYBYTES);
    }

    state = (crypto_onetimeauth_state *)lua_newuserdata(L,
      crypto_onetimeauth_statebytes());

    if(state == NULL) {
        return luaL_error(L,"out of memory");
    }

    if(crypto_onetimeauth_init(state,k) == -1) {
        lua_pop(L,1);
        return luaL_error(L,"crypto_onetimeauth_init error");
    }

    lua_pushvalue(L,lua_upvalueindex(1));
    lua_setmetatable(L,-2);

    return 1;
}

static int
ls_crypto_onetimeauth_update(lua_State *L) {
    crypto_onetimeauth_state *state = NULL;
    const unsigned char *m   = NULL;
    size_t mlen = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 parameters");
    }

    /* verify this is a state object */
    lua_pushvalue(L,lua_upvalueindex(1));
    lua_getmetatable(L,1);

    if(!ls_lua_equal(L,-2,-1)) {
        return luaL_error(L,"invalid userdata");
    }
    lua_pop(L,2);

    state = (crypto_onetimeauth_state *)lua_touserdata(L,1);
    m   = (const unsigned char *)lua_tolstring(L,2,&mlen);

    lua_pushboolean(L,crypto_onetimeauth_update(
      state,m,mlen) != -1);
    return 1;
}

static int
ls_crypto_onetimeauth_final(lua_State *L) {
    crypto_onetimeauth_state *state = NULL;
    unsigned char h[crypto_onetimeauth_BYTES];

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameters");
    }

    /* verify this is a state object */
    lua_pushvalue(L,lua_upvalueindex(1));
    lua_getmetatable(L,1);

    if(!ls_lua_equal(L,-2,-1)) {
        return luaL_error(L,"invalid userdata");
    }
    lua_pop(L,2);

    state = (crypto_onetimeauth_state *)lua_touserdata(L,1);

    if(crypto_onetimeauth_final(state,h) == -1) {
        return luaL_error(L,"crypto_onetimeauth_final error");
    }

    lua_pushlstring(L,(const char *)h,crypto_onetimeauth_BYTES);
    sodium_memzero(h,crypto_onetimeauth_BYTES);

    return 1;
}

static int
ls_crypto_onetimeauth_state__gc(lua_State *L) {
    crypto_onetimeauth_state *state = (crypto_onetimeauth_state *)lua_touserdata(L,1);
    sodium_memzero(state,crypto_onetimeauth_statebytes());
    return 0;
}

static const struct luaL_Reg ls_crypto_onetimeauth_functions[] = {
    LS_LUA_FUNC(crypto_onetimeauth),
    LS_LUA_FUNC(crypto_onetimeauth_verify),
    LS_LUA_FUNC(crypto_onetimeauth_keygen),
    { NULL, NULL },
};

static const struct luaL_Reg ls_crypto_onetimeauth_state_functions[] = {
    LS_LUA_FUNC(crypto_onetimeauth_init),
    LS_LUA_FUNC(crypto_onetimeauth_update),
    LS_LUA_FUNC(crypto_onetimeauth_final),
    { NULL, NULL },
};

static int
ls_crypto_onetimeauth_state_setup(lua_State *L) {
    /* create our metatable for crypto_onetimeauth_state */
    lua_newtable(L);
    lua_pushcclosure(L,ls_crypto_onetimeauth_state__gc,0);
    lua_setfield(L,-2,"__gc");

    /* table of methods */
    lua_newtable(L);
    lua_setfield(L,-2,"__index");

    /* top of stack is our metatable */
    /* push up copies of our module + metatable since setfuncs will pop metatable */
    lua_pushvalue(L,-2); /* module */
    lua_pushvalue(L,-2); /* metatable */
    luaL_setfuncs(L,ls_crypto_onetimeauth_state_functions,1);
    lua_pop(L,1); /* module (copy) */

    /* stack is now:
     *   table (our modules)
     *   table (our metatable)
     */

    lua_getfield(L,-1,"__index");
    /* module
     * metatable
     * __index
     */

    lua_getfield(L,-3,"crypto_onetimeauth_update");
    /* module
     * metatable
     * __index
     * function
     */
    lua_setfield(L,-2,"update");

    lua_getfield(L,-3,"crypto_onetimeauth_final");
    lua_setfield(L,-2,"final");

    /* module
     * metatable
     * __index
     */

    lua_pop(L,2);
    return 0;
}


static int
ls_crypto_onetimeauth_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_crypto_onetimeauth_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_crypto_onetimeauth_functions,0);
    ls_crypto_onetimeauth_state_setup(L);
    return 0;
}

static const luasodium_constant_t ls_crypto_scalarmult_constants[] = {
    LS_CONST(crypto_scalarmult_BYTES),
    LS_CONST(crypto_scalarmult_SCALARBYTES),
    { NULL, 0 },
};



static int
ls_crypto_scalarmult_base(lua_State *L) {
    unsigned char q[crypto_scalarmult_BYTES];
    const unsigned char *n = NULL;
    size_t n_len = 0;

    if(lua_isnoneornil(L,1)) {
        lua_pushliteral(L,"requires 1 argument");
        return lua_error(L);
    }

    n = (const unsigned char *)lua_tolstring(L,1,&n_len);
    if(n_len != crypto_scalarmult_SCALARBYTES) {
        return luaL_error(L,"wrong scalar length, expected: %d",
          crypto_scalarmult_SCALARBYTES);
    }

    if(crypto_scalarmult_base(q,n) == -1) {
        return luaL_error(L,"crypto_scalarmult_base error");
    }

    lua_pushlstring(L,(const char *)q,crypto_scalarmult_BYTES);
    sodium_memzero(q,crypto_scalarmult_BYTES);
    return 1;
}

static int
ls_crypto_scalarmult(lua_State *L) {
    unsigned char q[crypto_scalarmult_BYTES];
    const unsigned char *n = NULL;
    const unsigned char *p = NULL;
    size_t n_len = 0;
    size_t p_len = 0;

    if(lua_isnoneornil(L,2)) {
        lua_pushliteral(L,"requires 2 arguments");
        return lua_error(L);
    }

    n = (const unsigned char *)lua_tolstring(L,1,&n_len);
    if(n_len != crypto_scalarmult_SCALARBYTES) {
        return luaL_error(L,"wrong scalar length, expected: %d",
          crypto_scalarmult_SCALARBYTES);
    }

    p = (const unsigned char *)lua_tolstring(L,1,&p_len);
    if(p_len != crypto_scalarmult_BYTES) {
        return luaL_error(L,"wrong scalar length, expected: %d",
          crypto_scalarmult_BYTES);
    }

    if(crypto_scalarmult(q,n,p) == -1) {
        return luaL_error(L,"crypto_scalarmult error");
    }

    lua_pushlstring(L,(const char *)q,crypto_scalarmult_BYTES);
    sodium_memzero(q,crypto_scalarmult_BYTES);
    return 1;
}

static const luaL_Reg ls_crypto_scalarmult_functions[] = {
    LS_LUA_FUNC(crypto_scalarmult_base),
    LS_LUA_FUNC(crypto_scalarmult),
    { NULL }
};

static int
ls_crypto_scalarmult_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_crypto_scalarmult_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_crypto_scalarmult_functions,0);
    return 0;
}
static const luasodium_constant_t ls_crypto_secretbox_constants[] = {
    LS_CONST(crypto_secretbox_KEYBYTES),
    LS_CONST(crypto_secretbox_MACBYTES),
    LS_CONST(crypto_secretbox_NONCEBYTES),

    LS_CONST(crypto_secretbox_xsalsa20poly1305_KEYBYTES),
    LS_CONST(crypto_secretbox_xsalsa20poly1305_NONCEBYTES),
    LS_CONST(crypto_secretbox_xsalsa20poly1305_MACBYTES),

    LS_CONST(crypto_secretbox_xchacha20poly1305_KEYBYTES),
    LS_CONST(crypto_secretbox_xchacha20poly1305_NONCEBYTES),
    LS_CONST(crypto_secretbox_xchacha20poly1305_MACBYTES),

    LS_CONST(crypto_secretbox_ZEROBYTES),
    LS_CONST(crypto_secretbox_xsalsa20poly1305_ZEROBYTES),

    LS_CONST(crypto_secretbox_BOXZEROBYTES),
    LS_CONST(crypto_secretbox_xsalsa20poly1305_BOXZEROBYTES),
    { NULL, 0 },
};




/* crypto_secretbox_keygen() */
static int
ls_crypto_secretbox_keygen(lua_State *L) {
    unsigned char k[crypto_secretbox_KEYBYTES];
    crypto_secretbox_keygen(k);
    lua_pushlstring(L,(const char *)k,crypto_secretbox_KEYBYTES);
    sodium_memzero(k,crypto_secretbox_KEYBYTES);
    return 1;
}


/* crypto_secretbox, crypto_secretbox_xsalsa20poly1305, etc */
static int
ls_crypto_secretbox(lua_State *L) {
    unsigned char *c      = NULL;
    const unsigned char *m = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *key   = NULL;

    unsigned char *tmp_m   = NULL;

    size_t clen = 0;
    size_t mlen = 0;
    size_t noncelen = 0;
    size_t keylen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 parameters");
        return lua_error(L);
    }

    m = (const unsigned char *)lua_tolstring(L,1,&mlen);
    nonce = (const unsigned char *)lua_tolstring(L,2,&noncelen);
    key   = (const unsigned char *)lua_tolstring(L,3,&keylen);

    if(noncelen != crypto_secretbox_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_secretbox_NONCEBYTES);
    }

    if(keylen != crypto_secretbox_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",crypto_secretbox_KEYBYTES);
    }

    clen = mlen + crypto_secretbox_MACBYTES;

    tmp_m = (unsigned char *)lua_newuserdata(L,mlen + crypto_secretbox_ZEROBYTES);
    if(tmp_m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    c = (unsigned char *)lua_newuserdata(L,clen + crypto_secretbox_BOXZEROBYTES);
    if(c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,2);

    sodium_memzero(tmp_m,crypto_secretbox_ZEROBYTES);
    sodium_memzero(c,crypto_secretbox_BOXZEROBYTES);

    memcpy(&tmp_m[crypto_secretbox_ZEROBYTES],m,mlen);

    if(crypto_secretbox(c,tmp_m,mlen+crypto_secretbox_ZEROBYTES,nonce,key) == -1) {
        return luaL_error(L,"crypto_secretbox error");
    }

    lua_pushlstring(L,(const char *)&c[crypto_secretbox_BOXZEROBYTES],clen);
    sodium_memzero(tmp_m,mlen + crypto_secretbox_ZEROBYTES);
    sodium_memzero(c,clen + crypto_secretbox_BOXZEROBYTES);
    return 1;
}

/* crypto_secretbox_open, crypto_secretbox_xsalsa20poly1305_open, etc */
static int
ls_crypto_secretbox_open(lua_State *L) {
    unsigned char *m      = NULL;
    const unsigned char *c = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *key   = NULL;

    unsigned char *tmp_c   = NULL;

    size_t mlen = 0;
    size_t clen = 0;
    size_t noncelen = 0;
    size_t keylen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 parameters");
        return lua_error(L);
    }

    c = (const unsigned char *)lua_tolstring(L,1,&clen);
    nonce = (const unsigned char *)lua_tolstring(L,2,&noncelen);
    key   = (const unsigned char *)lua_tolstring(L,3,&keylen);

    if(clen <= crypto_secretbox_MACBYTES) {
        return luaL_error(L,"wrong mac size, expected at least: %d",crypto_secretbox_MACBYTES);
    }

    if(noncelen != crypto_secretbox_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_secretbox_NONCEBYTES);
    }

    if(keylen != crypto_secretbox_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",crypto_secretbox_KEYBYTES);
    }

    mlen = clen - crypto_secretbox_MACBYTES;

    tmp_c = (unsigned char *)lua_newuserdata(L,clen + crypto_secretbox_BOXZEROBYTES);
    if(tmp_c == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    m = (unsigned char *)lua_newuserdata(L,mlen + crypto_secretbox_ZEROBYTES);
    if(m == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,2);

    sodium_memzero(tmp_c,crypto_secretbox_BOXZEROBYTES);
    sodium_memzero(m,crypto_secretbox_ZEROBYTES);

    memcpy(&tmp_c[crypto_secretbox_BOXZEROBYTES],c,clen);

    if(crypto_secretbox_open(m,tmp_c,clen+crypto_secretbox_BOXZEROBYTES,nonce,key) == -1) {
        return luaL_error(L,"crypto_secretbox_open error");
    }

    lua_pushlstring(L,(const char *)&m[crypto_secretbox_ZEROBYTES],mlen);
    sodium_memzero(tmp_c,clen + crypto_secretbox_BOXZEROBYTES);
    sodium_memzero(m,mlen + crypto_secretbox_ZEROBYTES);
    return 1;
}

/* crypto_secretbox, crypto_secretbox_xsalsa20poly1305, etc */
static int
ls_crypto_secretbox_easy(lua_State *L) {
    unsigned char *output      = NULL;
    const unsigned char *input = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *key   = NULL;

    size_t outputlen = 0;
    size_t inputlen = 0;
    size_t noncelen = 0;
    size_t keylen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 parameters");
        return lua_error(L);
    }

    input = (const unsigned char *)lua_tolstring(L,1,&inputlen);
    nonce = (const unsigned char *)lua_tolstring(L,2,&noncelen);
    key   = (const unsigned char *)lua_tolstring(L,3,&keylen);

    if(noncelen != crypto_secretbox_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_secretbox_NONCEBYTES);
    }

    if(keylen != crypto_secretbox_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",crypto_secretbox_KEYBYTES);
    }

    outputlen = inputlen + crypto_secretbox_MACBYTES;

    output = (unsigned char *)lua_newuserdata(L,outputlen);
    if(output == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_secretbox_easy(output,input,inputlen,nonce,key) == -1) {
        return luaL_error(L,"crypto_secretbox_easy error");
    }

    lua_pushlstring(L,(const char *)output,outputlen);
    sodium_memzero(output,outputlen);
    return 1;
}

/* crypto_secretbox, crypto_secretbox_xsalsa20poly1305, etc */
static int
ls_crypto_secretbox_open_easy(lua_State *L) {
    unsigned char *output      = NULL;
    const unsigned char *input = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *key   = NULL;

    size_t outputlen = 0;
    size_t inputlen = 0;
    size_t noncelen = 0;
    size_t keylen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 parameters");
        return lua_error(L);
    }

    input = (const unsigned char *)lua_tolstring(L,1,&inputlen);
    nonce = (const unsigned char *)lua_tolstring(L,2,&noncelen);
    key   = (const unsigned char *)lua_tolstring(L,3,&keylen);

    if(inputlen <= crypto_secretbox_MACBYTES) {
        return luaL_error(L,"wrong mac size, expected at least: %d",crypto_secretbox_MACBYTES);
    }

    if(noncelen != crypto_secretbox_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_secretbox_NONCEBYTES);
    }

    if(keylen != crypto_secretbox_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",crypto_secretbox_KEYBYTES);
    }

    outputlen = inputlen - crypto_secretbox_MACBYTES;

    output = (unsigned char *)lua_newuserdata(L,outputlen);
    if(output == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_secretbox_open_easy(output,input,inputlen,nonce,key) == -1) {
        return luaL_error(L,"crypto_secretbox_open_easy error");
    }

    lua_pushlstring(L,(const char *)output,outputlen);
    sodium_memzero(output,outputlen);
    return 1;
}

/* crypto_secretbox_detached(message, nonce, key) */
static int
ls_crypto_secretbox_detached(lua_State *L) {
    unsigned char *output = NULL;
    unsigned char mac[crypto_secretbox_MACBYTES];
    const unsigned char *input = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *key = NULL;

    size_t inputlen = 0;
    size_t noncelen = 0;
    size_t keylen = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 parameters");
        return lua_error(L);
    }

    input = (const unsigned char *)lua_tolstring(L,1,&inputlen);
    nonce = (const unsigned char *)lua_tolstring(L,2,&noncelen);
    key   = (const unsigned char *)lua_tolstring(L,3,&keylen);

    if(noncelen != crypto_secretbox_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_secretbox_NONCEBYTES);
    }

    if(keylen != crypto_secretbox_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",crypto_secretbox_KEYBYTES);
    }

    output = lua_newuserdata(L,inputlen);
    if(output == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(crypto_secretbox_detached(output,mac,input,inputlen,nonce,key) == -1) {
        return luaL_error(L,"crypto_secretbox_detached error");
    }

    lua_pushlstring(L,(const char *)output,inputlen);
    lua_pushlstring(L,(const char *)mac,crypto_secretbox_MACBYTES);

    sodium_memzero(output,inputlen);
    sodium_memzero(mac,crypto_secretbox_MACBYTES);
    return 2;
}

/* crypto_secretbox_open_detached(cipher, mac, nonce, key) */
static int
ls_crypto_secretbox_open_detached(lua_State *L) {
    unsigned char *output = NULL;
    const unsigned char *input = NULL;
    const unsigned char *mac = NULL;
    const unsigned char *nonce = NULL;
    const unsigned char *key = NULL;

    size_t inputlen = 0;
    size_t noncelen = 0;
    size_t keylen = 0;
    size_t maclen = 0;

    if(lua_isnoneornil(L,4)) {
        lua_pushliteral(L,"requires 4 parameters");
        return lua_error(L);
    }

    input = (const unsigned char *)lua_tolstring(L,1,&inputlen);
    mac   = (const unsigned char *)lua_tolstring(L,2,&maclen);
    nonce = (const unsigned char *)lua_tolstring(L,3,&noncelen);
    key   = (const unsigned char *)lua_tolstring(L,4,&keylen);

    if(maclen != crypto_secretbox_MACBYTES) {
        return luaL_error(L,"wrong mac size, expected: %d",crypto_secretbox_MACBYTES);
    }

    if(noncelen != crypto_secretbox_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",crypto_secretbox_NONCEBYTES);
    }

    if(keylen != crypto_secretbox_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",crypto_secretbox_KEYBYTES);
    }

    output = lua_newuserdata(L,inputlen);
    if(output == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }

    lua_pop(L,1);

    if(crypto_secretbox_open_detached(output,input,mac,inputlen,nonce,key) == -1) {
        return luaL_error(L,"crypto_secretbox_open_detached error");
    }

    lua_pushlstring(L,(const char *)output,inputlen);
    sodium_memzero(output,inputlen);
    return 1;
}

static const struct luaL_Reg ls_crypto_secretbox_funcs[] = {
    LS_LUA_FUNC(crypto_secretbox_keygen),
    LS_LUA_FUNC(crypto_secretbox),
    LS_LUA_FUNC(crypto_secretbox_open),
    LS_LUA_FUNC(crypto_secretbox_easy),
    LS_LUA_FUNC(crypto_secretbox_open_easy),
    LS_LUA_FUNC(crypto_secretbox_detached),
    LS_LUA_FUNC(crypto_secretbox_open_detached),
    { NULL, NULL }
};

static int
ls_crypto_secretbox_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_crypto_secretbox_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_crypto_secretbox_funcs,0);
    return 0;
}
static const luasodium_constant_t ls_crypto_sign_constants[] = {
    LS_CONST(crypto_sign_PUBLICKEYBYTES),
    LS_CONST(crypto_sign_SECRETKEYBYTES),
    LS_CONST(crypto_sign_BYTES),
    LS_CONST(crypto_sign_SEEDBYTES),
    { NULL, 0 },
};




static int
ls_crypto_sign_keypair(lua_State *L) {
    unsigned char pk[crypto_sign_PUBLICKEYBYTES];
    unsigned char sk[crypto_sign_SECRETKEYBYTES];

    if(crypto_sign_keypair(pk,sk) == -1) {
        return luaL_error(L,"crypto_sign_keypair error");
    }

    lua_pushlstring(L,(const char *)pk,crypto_sign_PUBLICKEYBYTES);
    lua_pushlstring(L,(const char *)sk,crypto_sign_SECRETKEYBYTES);

    sodium_memzero(pk,crypto_sign_PUBLICKEYBYTES);
    sodium_memzero(sk,crypto_sign_SECRETKEYBYTES);

    return 2;
}

static int
ls_crypto_sign_seed_keypair(lua_State *L) {
    unsigned char pk[crypto_sign_PUBLICKEYBYTES];
    unsigned char sk[crypto_sign_SECRETKEYBYTES];
    const unsigned char *seed = NULL;
    size_t seedlen = 0;

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameter");
    }

    seed = (const unsigned char *)lua_tolstring(L,1,&seedlen);

    if(seedlen != crypto_sign_SEEDBYTES) {
        return luaL_error(L,"wrong seed size, expected: %d",
          crypto_sign_SEEDBYTES);
    }

    if(crypto_sign_seed_keypair(pk,sk,seed) == -1) {
        return luaL_error(L,"crypto_sign_keypair error");
    }

    lua_pushlstring(L,(const char *)pk,crypto_sign_PUBLICKEYBYTES);
    lua_pushlstring(L,(const char *)sk,crypto_sign_SECRETKEYBYTES);

    sodium_memzero(pk,crypto_sign_PUBLICKEYBYTES);
    sodium_memzero(sk,crypto_sign_SECRETKEYBYTES);

    return 2;
}

static int
ls_crypto_sign(lua_State *L) {
    unsigned char *sm = NULL;
    const unsigned char *m = NULL;
    const unsigned char *sk = NULL;
    unsigned long long smlen;
    size_t smlen_a = 0;
    size_t mlen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 parameters");
    }

    m  = (const unsigned char *)lua_tolstring(L,1,&mlen);
    sk = (const unsigned char *)lua_tolstring(L,2,&sklen);

    if(sklen != crypto_sign_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key size, expected: %d",
          crypto_sign_SECRETKEYBYTES);
    }

    smlen_a = mlen + crypto_sign_BYTES;

    sm = lua_newuserdata(L,smlen_a);
    if(sm == NULL) {
        return luaL_error(L,"out of memory");
    }
    lua_pop(L,1);

    if(crypto_sign(sm,&smlen,m,mlen,sk) == -1) {
        return luaL_error(L,"crypto_sign error");
    }
    lua_pushlstring(L,(const char *)sm,smlen);
    sodium_memzero(sm,smlen_a);
    return 1;
}

static int
ls_crypto_sign_open(lua_State *L) {
    unsigned char *m = NULL;
    const unsigned char *sm = NULL;
    const unsigned char *pk = NULL;
    unsigned long long mlen;
    size_t smlen = 0;
    size_t pklen = 0;
    int r = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 parameters");
    }

    sm = (const unsigned char *)lua_tolstring(L,1,&smlen);
    pk = (const unsigned char *)lua_tolstring(L,2,&pklen);

    if(pklen != crypto_sign_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key size, expected: %d",
          crypto_sign_PUBLICKEYBYTES);
    }

    m = lua_newuserdata(L,smlen);
    if(sm == NULL) {
        return luaL_error(L,"out of memory");
    }
    lua_pop(L,1);

    if(crypto_sign_open(m,&mlen,sm,smlen,pk) == 0) {
        lua_pushlstring(L,(const char *)m,mlen);
        r = 1;
    }
    sodium_memzero(m,smlen);
    return r;
}

static int
ls_crypto_sign_detached(lua_State *L) {
    unsigned char sig[crypto_sign_BYTES];
    const unsigned char *m = NULL;
    const unsigned char *sk = NULL;
    unsigned long long siglen;
    size_t mlen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 parameters");
    }

    m  = (const unsigned char *)lua_tolstring(L,1,&mlen);
    sk = (const unsigned char *)lua_tolstring(L,2,&sklen);

    if(sklen != crypto_sign_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key size, expected: %d",
          crypto_sign_SECRETKEYBYTES);
    }

    if(crypto_sign_detached(sig,&siglen,m,mlen,sk) == -1) {
        return luaL_error(L,"crypto_sign_detached error");
    }
    lua_pushlstring(L,(const char *)sig,siglen);
    sodium_memzero(sig,crypto_sign_BYTES);
    return 1;
}

static int
ls_crypto_sign_verify_detached(lua_State *L) {
    const unsigned char *sig = NULL;
    const unsigned char *m   = NULL;
    const unsigned char *pk  = NULL;
    size_t siglen = 0;
    size_t mlen = 0;
    size_t pklen = 0;

    if(lua_isnoneornil(L,3)) {
        return luaL_error(L,"requires 3 parameters");
    }

    sig = (const unsigned char *)lua_tolstring(L,1,&siglen);
    m   = (const unsigned char *)lua_tolstring(L,2,&mlen);
    pk  = (const unsigned char *)lua_tolstring(L,3,&pklen);

    if(pklen != crypto_sign_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key size, expected: %d",
          crypto_sign_PUBLICKEYBYTES);
    }

    lua_pushboolean(L,crypto_sign_verify_detached(
      sig,m,mlen,pk) == 0);
    return 1;
}

static int
ls_crypto_sign_init(lua_State *L) {
    crypto_sign_state *state = NULL;

    state = (crypto_sign_state *)lua_newuserdata(L,
      crypto_sign_statebytes());
    if(crypto_sign_init(state) == -1) {
        lua_pop(L,1);
        return luaL_error(L,"crypto_sign_init error");
    }

    lua_pushvalue(L,lua_upvalueindex(1));
    lua_setmetatable(L,-2);

    return 1;
}

static int
ls_crypto_sign_update(lua_State *L) {
    crypto_sign_state *state = NULL;
    const unsigned char *m   = NULL;
    size_t mlen = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 parameters");
    }

    /* verify this is a state object */
    lua_pushvalue(L,lua_upvalueindex(1));
    lua_getmetatable(L,1);

    if(!ls_lua_equal(L,-2,-1)) {
        return luaL_error(L,"invalid userdata");
    }
    lua_pop(L,2);

    state = (crypto_sign_state *)lua_touserdata(L,1);
    m   = (const unsigned char *)lua_tolstring(L,2,&mlen);

    lua_pushboolean(L,crypto_sign_update(
      state,m,mlen) != -1);
    return 1;
}

static int
ls_crypto_sign_final_create(lua_State *L) {
    crypto_sign_state *state = NULL;
    unsigned char sig[crypto_sign_BYTES];
    const unsigned char *sk   = NULL;
    unsigned long long siglen = 0;
    size_t sklen = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 parameters");
    }

    /* verify this is a state object */
    lua_pushvalue(L,lua_upvalueindex(1));
    lua_getmetatable(L,1);

    if(!ls_lua_equal(L,-2,-1)) {
        return luaL_error(L,"invalid userdata");
    }
    lua_pop(L,2);

    state = (crypto_sign_state *)lua_touserdata(L,1);
    sk  = (const unsigned char *)lua_tolstring(L,2,&sklen);

    if(sklen != crypto_sign_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key size, expected: %d",
          crypto_sign_SECRETKEYBYTES);
    }

    if(crypto_sign_final_create(state,sig,&siglen,sk) == -1) {
        return luaL_error(L,"crypto_sign_final_create error");
    }

    lua_pushlstring(L,(const char *)sig,siglen);
    sodium_memzero(sig,siglen);
    return 1;
}

static int
ls_crypto_sign_final_verify(lua_State *L) {
    crypto_sign_state *state  = NULL;
    const unsigned char *sig  = NULL;
    const unsigned char *pk   = NULL;
    size_t siglen = 0;
    size_t pklen = 0;

    if(lua_isnoneornil(L,3)) {
        return luaL_error(L,"requires 3 parameters");
    }

    /* verify this is a state object */
    lua_pushvalue(L,lua_upvalueindex(1));
    lua_getmetatable(L,1);

    if(!ls_lua_equal(L,-2,-1)) {
        return luaL_error(L,"invalid userdata");
    }
    lua_pop(L,2);

    state = (crypto_sign_state *)lua_touserdata(L,1);
    sig = (const unsigned char *)lua_tolstring(L,2,&siglen);
    pk  = (const unsigned char *)lua_tolstring(L,3,&pklen);

    if(pklen != crypto_sign_PUBLICKEYBYTES) {
        return luaL_error(L,"wrong public key size, expected: %d",
          crypto_sign_PUBLICKEYBYTES);
    }

    lua_pushboolean(L,crypto_sign_final_verify(
      state,sig,pk) == 0);
    return 1;
}

static int
ls_crypto_sign_ed25519_sk_to_seed(lua_State *L) {
    unsigned char seed[crypto_sign_SEEDBYTES];
    const unsigned char *sk = NULL;
    size_t sklen = 0;

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameter");
    }

    sk  = (const unsigned char *)lua_tolstring(L,1,&sklen);

    if(sklen != crypto_sign_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key size, expected: %d",
          crypto_sign_SECRETKEYBYTES);
    }

    if(crypto_sign_ed25519_sk_to_seed(seed,sk) == -1) {
        return luaL_error(L,"crypto_sign_ed25519_sk_to_seed error");
    }
    lua_pushlstring(L,(const char *)seed,crypto_sign_SEEDBYTES);
    sodium_memzero(seed,crypto_sign_SEEDBYTES);
    return 1;
}

static int
ls_crypto_sign_ed25519_sk_to_pk(lua_State *L) {
    unsigned char pk[crypto_sign_PUBLICKEYBYTES];
    const unsigned char *sk = NULL;
    size_t sklen = 0;

    if(lua_isnoneornil(L,1)) {
        return luaL_error(L,"requires 1 parameter");
    }

    sk  = (const unsigned char *)lua_tolstring(L,1,&sklen);

    if(sklen != crypto_sign_SECRETKEYBYTES) {
        return luaL_error(L,"wrong secret key size, expected: %d",
          crypto_sign_SECRETKEYBYTES);
    }

    if(crypto_sign_ed25519_sk_to_pk(pk,sk) == -1) {
        return luaL_error(L,"crypto_sign_ed25519_sk_to_pk error");
    }
    lua_pushlstring(L,(const char *)pk,crypto_sign_PUBLICKEYBYTES);
    sodium_memzero(pk,crypto_sign_PUBLICKEYBYTES);
    return 1;
}

static int
ls_crypto_sign_state__gc(lua_State *L) {
    crypto_sign_state *state = (crypto_sign_state *)lua_touserdata(L,1);
    sodium_memzero(state,crypto_sign_statebytes());
    return 0;
}

static const struct luaL_Reg ls_crypto_sign_functions[] = {
    LS_LUA_FUNC(crypto_sign_keypair),
    LS_LUA_FUNC(crypto_sign_seed_keypair),
    LS_LUA_FUNC(crypto_sign),
    LS_LUA_FUNC(crypto_sign_open),
    LS_LUA_FUNC(crypto_sign_detached),
    LS_LUA_FUNC(crypto_sign_verify_detached),
    LS_LUA_FUNC(crypto_sign_ed25519_sk_to_seed),
    LS_LUA_FUNC(crypto_sign_ed25519_sk_to_pk),
    { NULL, NULL },
};

static const struct luaL_Reg ls_crypto_sign_state_functions[] = {
    LS_LUA_FUNC(crypto_sign_init),
    LS_LUA_FUNC(crypto_sign_update),
    LS_LUA_FUNC(crypto_sign_final_create),
    LS_LUA_FUNC(crypto_sign_final_verify),
    { NULL, NULL },
};

static int
ls_crypto_sign_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_crypto_sign_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_crypto_sign_functions,0);

    /* create our metatable for crypto_sign_state */
    lua_newtable(L);
    lua_pushcclosure(L,ls_crypto_sign_state__gc,0);
    lua_setfield(L,-2,"__gc");

    /* table of methods */
    lua_newtable(L);
    lua_setfield(L,-2,"__index");

    /* top of stack is our metatable */
    /* push up copies of our module + metatable since setfuncs will pop metatable */
    lua_pushvalue(L,-2); /* module */
    lua_pushvalue(L,-2); /* metatable */
    luaL_setfuncs(L,ls_crypto_sign_state_functions,1);
    lua_pop(L,1); /* module (copy) */

    /* stack is now:
     *   table (our modules)
     *   table (our metatable)
     */

    lua_getfield(L,-1,"__index");
    /* module
     * metatable
     * __index
     */

    lua_getfield(L,-3,"crypto_sign_update");
    /* module
     * metatable
     * __index
     * function
     */
    lua_setfield(L,-2,"update");

    lua_getfield(L,-3,"crypto_sign_final_create");
    lua_setfield(L,-2,"final_create");

    lua_getfield(L,-3,"crypto_sign_final_verify");
    lua_setfield(L,-2,"final_verify");

    /* module
     * metatable
     * __index
     */

    lua_pop(L,2);

    return 0;
}

static const luasodium_constant_t ls_crypto_stream_constants[] = {
    LS_CONST(crypto_stream_KEYBYTES),
    LS_CONST(crypto_stream_NONCEBYTES),
    { NULL, 0 },
};


static int
ls_crypto_stream(lua_State *L) {
    unsigned char *c = NULL;
    const unsigned char *n = NULL;
    const unsigned char *k = NULL;
    size_t clen = 0;
    size_t nlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,3)) {
        return luaL_error(L,"requires 3 parameter");
    }

    clen = lua_tointeger(L,1);
    n = (const unsigned char *)lua_tolstring(L,2,&nlen);
    k = (const unsigned char *)lua_tolstring(L,3,&klen);

    if(nlen != crypto_stream_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",
          crypto_stream_NONCEBYTES);
    }

    if(klen != crypto_stream_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",
          crypto_stream_KEYBYTES);
    }

    c = lua_newuserdata(L,clen);
    if(c == NULL) {
        return luaL_error(L,"out of memory");
    }
    lua_pop(L,1);

    if(crypto_stream(c,clen,n,k) == -1) {
        return luaL_error(L,"crypto_stream error");
    }

    lua_pushlstring(L,(const char *)c,clen);
    sodium_memzero(c,clen);
    return 1;
}

static int
ls_crypto_stream_xor(lua_State *L) {
    unsigned char *c = NULL;
    const unsigned char *m = NULL;
    const unsigned char *n = NULL;
    const unsigned char *k = NULL;
    size_t mlen = 0;
    size_t nlen = 0;
    size_t klen = 0;

    if(lua_isnoneornil(L,3)) {
        return luaL_error(L,"requires 3 parameter");
    }

    m = (const unsigned char *)lua_tolstring(L,1,&mlen);
    n = (const unsigned char *)lua_tolstring(L,2,&nlen);
    k = (const unsigned char *)lua_tolstring(L,3,&klen);

    if(nlen != crypto_stream_NONCEBYTES) {
        return luaL_error(L,"wrong nonce size, expected: %d",
          crypto_stream_NONCEBYTES);
    }

    if(klen != crypto_stream_KEYBYTES) {
        return luaL_error(L,"wrong key size, expected: %d",
          crypto_stream_KEYBYTES);
    }

    c = lua_newuserdata(L,mlen);
    if(c == NULL) {
        return luaL_error(L,"out of memory");
    }
    lua_pop(L,1);

    if(crypto_stream_xor(c,m,mlen,n,k) == -1) {
        return luaL_error(L,"crypto_stream_xor error");
    }

    lua_pushlstring(L,(const char *)c,mlen);
    sodium_memzero(c,mlen);
    return 1;
}

static int
ls_crypto_stream_keygen(lua_State *L) {
    unsigned char k[crypto_stream_KEYBYTES];

    crypto_stream_keygen(k);

    lua_pushlstring(L,(const char *)k,crypto_stream_KEYBYTES);
    sodium_memzero(k,crypto_stream_KEYBYTES);
    return 1;
}

static const struct luaL_Reg ls_crypto_stream_functions[] = {
    LS_LUA_FUNC(crypto_stream),
    LS_LUA_FUNC(crypto_stream_xor),
    LS_LUA_FUNC(crypto_stream_keygen),
    { NULL, NULL },
};

static int
ls_crypto_stream_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_crypto_stream_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_crypto_stream_functions,0);
    return 0;
}

static const luasodium_constant_t ls_crypto_verify_constants[] = {
    { NULL, 0 },
};



typedef int (*ls_verify_func_ptr)(const unsigned char *x, const unsigned char *y);
typedef struct ls_crypto_verify_def_s {
    const char *name;
    ls_verify_func_ptr func;
    size_t len;
} ls_crypto_verify_def;

#define LS_CRYPTO_VERIFY(x) { "crypto_verify_" #x, crypto_verify_ ## x, x }

static int
ls_crypto_verify(lua_State *L) {
    const unsigned char *x = NULL;
    const unsigned char *y = NULL;
    size_t xlen = 0;
    size_t ylen = 0;

    ls_verify_func_ptr f = NULL;
    size_t len = 0;

    if(lua_isnoneornil(L,2)) {
        return luaL_error(L,"requires 2 parameters");
    }

    f = lua_touserdata(L,lua_upvalueindex(1));
    len = lua_tointeger(L,lua_upvalueindex(2));

    x = (const unsigned char *)lua_tolstring(L,1,&xlen);
    y = (const unsigned char *)lua_tolstring(L,2,&ylen);

    if(xlen != len || ylen != len) {
        return luaL_error(L,"incorrect string size, expected: %d",
          len);
    }

    lua_pushboolean(L,f(x,y) == 0);
    return 1;
}

static const ls_crypto_verify_def ls_crypto_verify_defs[] = {
    LS_CRYPTO_VERIFY(16),
    LS_CRYPTO_VERIFY(32),
    { NULL }
};

static int
ls_crypto_verify_core_setup(lua_State *L) {
    const ls_crypto_verify_def *d = ls_crypto_verify_defs;
    luasodium_set_constants(L,ls_crypto_verify_constants,lua_gettop(L));

    for(; d->name != NULL; d++) {
        lua_pushlightuserdata(L,d->func);
        lua_pushinteger(L,d->len);
        lua_pushcclosure(L,ls_crypto_verify,2);
        lua_setfield(L,-2,d->name);
    }
    return 0;
}



static const luasodium_constant_t ls_randombytes_constants[] = {
    LS_CONST(randombytes_SEEDBYTES),
    { NULL, 0 },
};



static int
ls_randombytes_random(lua_State *L) {
    lua_pushinteger(L,randombytes_random());
    return 1;
}

static int
ls_randombytes_uniform(lua_State *L) {
    if(!lua_isnumber(L,1)) {
        lua_pushliteral(L,"missing number argument");
        return lua_error(L);
    }

    lua_pushinteger(L,randombytes_uniform(lua_tointeger(L,1)));
    return 1;
}

static int
ls_randombytes_buf(lua_State *L) {
    char *buf = NULL;
    size_t s = 0;

    if(!lua_isnumber(L,1)) {
        lua_pushliteral(L,"missing number argument");
        return lua_error(L);
    }

    s = lua_tointeger(L,1);
    buf = lua_newuserdata(L,s);
    if(buf == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    randombytes_buf(buf,s);
    lua_pushlstring(L,buf,s);
    sodium_memzero(buf,s);
    return 1;
}

static int
ls_randombytes_buf_deterministic(lua_State *L) {
    const char *seed = NULL;
    size_t seed_len;
    char *buf = NULL;
    size_t s = 0;

    if(lua_isnoneornil(L,2)) {
        lua_pushliteral(L,"requires 2 arguments");
        return lua_error(L);
    }

    s = lua_tointeger(L,1);
    seed = lua_tolstring(L,2,&seed_len);

    if(seed_len != randombytes_SEEDBYTES) {
        lua_pushliteral(L,"wrong seed length");
        return lua_error(L);
    }

    buf = lua_newuserdata(L,s);
    if(buf == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    randombytes_buf_deterministic(buf,s,(const unsigned char *)seed);

    lua_pushlstring(L,buf,s);
    sodium_memzero(buf,s);

    return 1;
}

static int
ls_randombytes_close(lua_State *L) {
    lua_pushboolean(L,randombytes_close() == 0);
    return 1;
}

static int
ls_randombytes_stir(lua_State *L) {
    (void)L;
    randombytes_stir();
    return 0;
}

static const struct luaL_Reg ls_randombytes_functions[] = {
    LS_LUA_FUNC(randombytes_random),
    LS_LUA_FUNC(randombytes_uniform),
    LS_LUA_FUNC(randombytes_buf),
    LS_LUA_FUNC(randombytes_buf_deterministic),
    LS_LUA_FUNC(randombytes_close),
    LS_LUA_FUNC(randombytes_stir),
    { NULL, NULL },
};

static int
ls_randombytes_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_randombytes_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_randombytes_functions,0);
    return 0;
}

static const luasodium_constant_t ls_utils_constants[] = {
    LS_CONST(sodium_base64_VARIANT_ORIGINAL),
    LS_CONST(sodium_base64_VARIANT_ORIGINAL_NO_PADDING),
    LS_CONST(sodium_base64_VARIANT_URLSAFE),
    LS_CONST(sodium_base64_VARIANT_URLSAFE_NO_PADDING),
    { NULL, 0 },
};



static int
ls_sodium_init(lua_State *L) {
    if(sodium_init() == -1) {
        lua_pushliteral(L,"sodium_init error");
        return lua_error(L);
    }
    lua_pushboolean(L,1);
    return 1;
}

static int
ls_sodium_memcmp(lua_State *L) {
    const char *b1 = NULL;
    const char *b2 = NULL;
    size_t len = 0;

    if(lua_isnoneornil(L,3)) {
        lua_pushliteral(L,"requires 3 arguments");
        return lua_error(L);
    }

    b1 = lua_tolstring(L,1,NULL);
    b2 = lua_tolstring(L,2,NULL);
    len = lua_tointeger(L,3);

    lua_pushboolean(L,sodium_memcmp(
      b1,
      b2,
      len) == 0);
    return 1;
}

/* luasodium.bin2hex(bin) */
static int
ls_sodium_bin2hex(lua_State *L) {
    const char *bin = NULL;
    char *hex = NULL;
    size_t bin_len = 0;
    size_t hex_len = 0;

    if(lua_isnoneornil(L,1)) {
        lua_pushliteral(L,"requires 1 argument");
        return lua_error(L);
    }

    bin = lua_tolstring(L,1,&bin_len);
    hex_len = (bin_len * 2);
    hex = lua_newuserdata(L,hex_len + 1);
    if(hex == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);
    sodium_bin2hex(hex,hex_len+1,(const unsigned char *)bin,bin_len);
    lua_pushstring(L,hex);
    sodium_memzero(hex,hex_len + 1);
    return 1;
}

/* luasodium.hex2bin(hex, [ignore]) */
static int
ls_sodium_hex2bin(lua_State *L) {
    const char *hex = NULL;
    const char *hex_end = NULL;
    const char *ignore = NULL;
    unsigned char *bin = NULL;
    int r = 1;

    size_t hex_len = 0;
    size_t bin_len = 0;
    size_t out_bin_len = 0;

    if(lua_isnoneornil(L,1)) {
        lua_pushliteral(L,"requires 1 argument");
        return lua_error(L);
    }

    hex = lua_tolstring(L,1,&hex_len);

    bin_len = hex_len / 2;
    if(hex_len % 2 != 0) {
        bin_len++;
    }

    if(lua_isstring(L,2)) {
        ignore = lua_tostring(L,2);
    }

    bin = lua_newuserdata(L,bin_len);
    if(bin == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(sodium_hex2bin(
        bin,bin_len,
        hex,hex_len,
        ignore, &out_bin_len,
        &hex_end) != 0) {
        lua_pushliteral(L,"error in hex2bin");
        return lua_error(L);
    }

    lua_pushlstring(L,(const char *)bin,out_bin_len);
    if(hex_end < hex + hex_len) {
        lua_pushlstring(L,hex_end,(hex + hex_len) - hex_end);
        r = 2;
    }
    sodium_memzero(bin,bin_len);
    return r;
}

/* luasodium.bin2base64(bin, variant) */
static int
ls_sodium_bin2base64(lua_State *L) {
    const char *bin = NULL;
    char *b64 = NULL;
    size_t bin_len = 0;
    size_t b64_len = 0;
    lua_Integer variant = 0;

    if(lua_isnoneornil(L,2)) {
        lua_pushliteral(L,"requires 2 argument");
        return lua_error(L);
    }

    bin = lua_tolstring(L,1,&bin_len);
    variant = lua_tointeger(L,2);

    switch(variant) {
        case sodium_base64_VARIANT_ORIGINAL: break;
        case sodium_base64_VARIANT_ORIGINAL_NO_PADDING: break;
        case sodium_base64_VARIANT_URLSAFE: break;
        case sodium_base64_VARIANT_URLSAFE_NO_PADDING: break;
        default: {
            lua_pushliteral(L,"unknown base64 variant");
            return lua_error(L);
        }
    }

    b64_len = sodium_base64_encoded_len(bin_len,variant);
    b64 = lua_newuserdata(L,b64_len);
    if(b64 == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    sodium_bin2base64(b64,b64_len,
      (const unsigned char *)bin, bin_len,
      variant);
    lua_pushstring(L,b64);
    sodium_memzero(b64,b64_len);
    return 1;
}

/* luasodium.base642bin(base64, variant, [ignore]) */
static int
ls_sodium_base642bin(lua_State *L) {
    const char *base64 = NULL;
    const char *base64_end = NULL;
    const char *ignore = NULL;
    unsigned char *bin = NULL;
    lua_Integer variant = 0;
    int r = 1;

    size_t base64_len;
    size_t bin_len;
    size_t out_bin_len;

    if(lua_isnoneornil(L,2)) {
        lua_pushliteral(L,"requires 2 arguments");
        return lua_error(L);
    }

    base64 = lua_tolstring(L,1,&base64_len);
    variant = lua_tointeger(L,2);

    switch(variant) {
        case sodium_base64_VARIANT_ORIGINAL: break;
        case sodium_base64_VARIANT_ORIGINAL_NO_PADDING: break;
        case sodium_base64_VARIANT_URLSAFE: break;
        case sodium_base64_VARIANT_URLSAFE_NO_PADDING: break;
        default: {
            lua_pushliteral(L,"unknown base64 variant");
            return lua_error(L);
        }
    }

    /* this is technicallly too many bytes but whatever */
    bin_len = base64_len;

    if(lua_isstring(L,3)) {
        ignore = lua_tostring(L,4);
    }

    bin = lua_newuserdata(L,bin_len);
    if(bin == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    if(sodium_base642bin(
        bin,bin_len,
        base64,base64_len,
        ignore, &out_bin_len,
        &base64_end,variant) != 0) {
        lua_pushliteral(L,"error in base642bin");
        return lua_error(L);
    }

    lua_pushlstring(L,(const char *)bin,out_bin_len);
    if(base64_end < base64 + base64_len) {
        lua_pushlstring(L,base64_end,(base64 + base64_len) - base64_end);
        r = 2;
    }
    sodium_memzero(bin,bin_len);
    return r;
}

static int
ls_sodium_increment(lua_State *L) {
    const char *n = NULL;
    char *r = NULL;
    size_t nlen = 0;

    n = lua_tolstring(L,1,&nlen);
    r = lua_newuserdata(L,nlen);
    if(r == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    memcpy(r,n,nlen);

    sodium_increment((unsigned char *)r,nlen);
    lua_pushlstring(L,r,nlen);
    sodium_memzero(r,nlen);
    return 1;
}

static int
ls_sodium_add(lua_State *L) {
    const char *a = NULL;
    const char *b = NULL;
    char *r = NULL;
    size_t alen = 0;
    size_t blen = 0;

    a = lua_tolstring(L,1,&alen);
    b = lua_tolstring(L,2,&blen);

    if(alen != blen) {
        lua_pushliteral(L,"mismatched data sizes");
        return lua_error(L);
    }

    r = lua_newuserdata(L,alen);
    if(r == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    memcpy(r,a,alen);
    sodium_add((unsigned char *)r,(const unsigned char *)b,alen);
    lua_pushlstring(L,r,alen);
    sodium_memzero(r,alen);
    return 1;
}

static int
ls_sodium_sub(lua_State *L) {
    const char *a = NULL;
    const char *b = NULL;
    char *r = NULL;
    size_t alen = 0;
    size_t blen = 0;

    a = lua_tolstring(L,1,&alen);
    b = lua_tolstring(L,2,&blen);

    if(alen != blen) {
        lua_pushliteral(L,"mismatched data sizes");
        return lua_error(L);
    }

    r = lua_newuserdata(L,alen);
    if(r == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);

    memcpy(r,a,alen);
    sodium_sub((unsigned char *)r,(const unsigned char *)b,alen);
    lua_pushlstring(L,r,alen);
    sodium_memzero(r,alen);
    return 1;
}

static int
ls_sodium_compare(lua_State *L) {
    const char *a = NULL;
    const char *b = NULL;
    size_t alen = 0;
    size_t blen = 0;

    a = lua_tolstring(L,1,&alen);
    b = lua_tolstring(L,2,&blen);

    if(alen != blen) {
        lua_pushliteral(L,"mismatched data sizes");
        return lua_error(L);
    }

    lua_pushinteger(L,sodium_compare((const void *)a,(const void *)b,alen));
    return 1;
}

static int
ls_sodium_is_zero(lua_State *L) {
    const char *n = NULL;
    size_t nlen = 0;

    if(lua_isnoneornil(L,1)) {
        lua_pushliteral(L,"1 argument");
        return lua_error(L);
    }

    n = lua_tolstring(L,1,&nlen);
    lua_pushboolean(L,sodium_is_zero((const unsigned char *)n,nlen) == 1);
    return 1;
}

static int
ls_sodium_pad(lua_State *L) {
    const char *n = NULL;
    char *r = NULL;
    size_t nlen = 0;
    size_t blocksize = 0;
    size_t rounded = 0;
    size_t outlen = 0;
    size_t rem = 0;

    if(lua_isnoneornil(L,2)) {
        lua_pushliteral(L,"requires 2 arguments");
        return lua_error(L);
    }

    n = lua_tolstring(L,1,&nlen);
    blocksize = lua_tointeger(L,2);

    rem = nlen % blocksize;
    rounded = nlen + (blocksize - rem);

    r = lua_newuserdata(L,rounded);
    if(r == NULL) {
        lua_pushliteral(L,"out of memory");
        return lua_error(L);
    }
    lua_pop(L,1);
    memcpy(r,n,nlen);

    if(sodium_pad(&outlen,(unsigned char *)r,
        nlen,blocksize,rounded) != 0) {
        sodium_memzero(r,rounded);
        lua_pushliteral(L,"sodium_pad error");
        return lua_error(L);
    }
    lua_pushlstring(L,r,outlen);
    sodium_memzero(r,rounded);
    return 1;
}

static int
ls_sodium_unpad(lua_State *L) {
    const char *n = NULL;
    size_t nlen = 0;
    size_t blocksize = 0;
    size_t outlen = 0;

    n = lua_tolstring(L,1,&nlen);
    blocksize = lua_tointeger(L,2);

    if(sodium_unpad(&outlen,(const unsigned char *)n,
        nlen,blocksize) != 0) {
        lua_pushliteral(L,"sodium_unpad error");
        return lua_error(L);
    }
    lua_pushlstring(L,n,outlen);
    return 1;
}

static const struct luaL_Reg ls_utils_functions[] = {
    LS_LUA_FUNC(sodium_init),
    LS_LUA_FUNC(sodium_memcmp),
    LS_LUA_FUNC(sodium_bin2hex),
    LS_LUA_FUNC(sodium_hex2bin),
    LS_LUA_FUNC(sodium_bin2base64),
    LS_LUA_FUNC(sodium_base642bin),
    LS_LUA_FUNC(sodium_increment),
    LS_LUA_FUNC(sodium_add),
    LS_LUA_FUNC(sodium_sub),
    LS_LUA_FUNC(sodium_compare),
    LS_LUA_FUNC(sodium_is_zero),
    LS_LUA_FUNC(sodium_pad),
    LS_LUA_FUNC(sodium_unpad),
    { NULL, NULL },
};

static int
ls_utils_core_setup(lua_State *L) {
    luasodium_set_constants(L,ls_utils_constants,lua_gettop(L));
    luaL_setfuncs(L,ls_utils_functions,0);
    return 0;
}

static const char ls_version_ffi_implementation[1039] = {
	0x6c, 0x6f, 0x63, 0x61, 0x6c, 0x20, 0x5f, 0x56, 0x45, 0x52, 0x53, 
	0x49, 0x4f, 0x4e, 0x20, 0x3d, 0x20, 0x27, 0x31, 0x2e, 0x30, 0x2e, 
	0x30, 0x27, 0x0a, 0x0a, 0x72, 0x65, 0x74, 0x75, 0x72, 0x6e, 0x20, 
	0x66, 0x75, 0x6e, 0x63, 0x74, 0x69, 0x6f, 0x6e, 0x28, 0x6c, 0x69, 
	0x62, 0x73, 0x29, 0x0a, 0x20, 0x20, 0x6c, 0x6f, 0x63, 0x61, 0x6c, 
	0x20, 0x4d, 0x20, 0x3d, 0x20, 0x7b, 0x0a, 0x20, 0x20, 0x20, 0x20, 
	0x5f, 0x56, 0x45, 0x52, 0x53, 0x49, 0x4f, 0x4e, 0x20, 0x3d, 0x20, 
	0x5f, 0x56, 0x45, 0x52, 0x53, 0x49, 0x4f, 0x4e, 0x0a, 0x20, 0x20, 
	0x7d, 0x0a, 0x0a, 0x20, 0x20, 0x2d, 0x2d, 0x20, 0x74, 0x68, 0x69, 
	0x73, 0x20, 0x6d, 0x61, 0x79, 0x20, 0x62, 0x65, 0x20, 0x72, 0x75, 
	0x6e, 0x6e, 0x69, 0x6e, 0x67, 0x20, 0x66, 0x72, 0x6f, 0x6d, 0x20, 
	0x6d, 0x79, 0x20, 0x67, 0x65, 0x6e, 0x65, 0x72, 0x69, 0x63, 0x20, 
	0x27, 0x73, 0x68, 0x6f, 0x77, 0x20, 0x76, 0x65, 0x72, 0x73, 0x69, 
	0x6f, 0x6e, 0x27, 0x20, 0x73, 0x63, 0x72, 0x69, 0x70, 0x74, 0x0a, 
	0x20, 0x20, 0x2d, 0x2d, 0x20, 0x2a, 0x6f, 0x72, 0x2a, 0x20, 0x66, 
	0x72, 0x6f, 0x6d, 0x20, 0x74, 0x68, 0x65, 0x20, 0x72, 0x65, 0x67, 
	0x75, 0x6c, 0x61, 0x72, 0x20, 0x43, 0x20, 0x41, 0x50, 0x49, 0x2e, 
	0x20, 0x69, 0x66, 0x20, 0x6c, 0x69, 0x62, 0x73, 0x20, 0x69, 0x73, 
	0x20, 0x65, 0x6d, 0x70, 0x74, 0x79, 0x2c, 0x20, 0x77, 0x65, 0x27, 
	0x72, 0x65, 0x20, 0x6e, 0x6f, 0x74, 0x20, 0x69, 0x6e, 0x0a, 0x20, 
	0x20, 0x2d, 0x2d, 0x20, 0x61, 0x6e, 0x20, 0x46, 0x46, 0x49, 0x20, 
	0x65, 0x6e, 0x76, 0x69, 0x72, 0x6f, 0x6e, 0x6d, 0x65, 0x6e, 0x74, 
	0x2c, 0x20, 0x73, 0x6f, 0x20, 0x72, 0x65, 0x74, 0x75, 0x72, 0x6e, 
	0x2e, 0x0a, 0x20, 0x20, 0x69, 0x66, 0x20, 0x6e, 0x6f, 0x74, 0x20, 
	0x6c, 0x69, 0x62, 0x73, 0x20, 0x74, 0x68, 0x65, 0x6e, 0x20, 0x72, 
	0x65, 0x74, 0x75, 0x72, 0x6e, 0x20, 0x4d, 0x20, 0x65, 0x6e, 0x64, 
	0x0a, 0x0a, 0x20, 0x20, 0x6c, 0x6f, 0x63, 0x61, 0x6c, 0x20, 0x66, 
	0x66, 0x69, 0x20, 0x3d, 0x20, 0x72, 0x65, 0x71, 0x75, 0x69, 0x72, 
	0x65, 0x27, 0x66, 0x66, 0x69, 0x27, 0x0a, 0x0a, 0x20, 0x20, 0x6c, 
	0x6f, 0x63, 0x61, 0x6c, 0x20, 0x66, 0x75, 0x6e, 0x63, 0x74, 0x69, 
	0x6f, 0x6e, 0x20, 0x6c, 0x73, 0x5f, 0x73, 0x6f, 0x64, 0x69, 0x75, 
	0x6d, 0x5f, 0x76, 0x65, 0x72, 0x73, 0x69, 0x6f, 0x6e, 0x5f, 0x73, 
	0x74, 0x72, 0x69, 0x6e, 0x67, 0x28, 0x29, 0x0a, 0x20, 0x20, 0x20, 
	0x20, 0x72, 0x65, 0x74, 0x75, 0x72, 0x6e, 0x20, 0x66, 0x66, 0x69, 
	0x2e, 0x73, 0x74, 0x72, 0x69, 0x6e, 0x67, 0x28, 0x6c, 0x69, 0x62, 
	0x73, 0x2e, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x2e, 0x73, 0x6f, 
	0x64, 0x69, 0x75, 0x6d, 0x5f, 0x76, 0x65, 0x72, 0x73, 0x69, 0x6f, 
	0x6e, 0x5f, 0x73, 0x74, 0x72, 0x69, 0x6e, 0x67, 0x28, 0x29, 0x29, 
	0x0a, 0x20, 0x20, 0x65, 0x6e, 0x64, 0x0a, 0x0a, 0x20, 0x20, 0x6c, 
	0x6f, 0x63, 0x61, 0x6c, 0x20, 0x66, 0x75, 0x6e, 0x63, 0x74, 0x69, 
	0x6f, 0x6e, 0x20, 0x6c, 0x73, 0x5f, 0x73, 0x6f, 0x64, 0x69, 0x75, 
	0x6d, 0x5f, 0x6c, 0x69, 0x62, 0x72, 0x61, 0x72, 0x79, 0x5f, 0x76, 
	0x65, 0x72, 0x73, 0x69, 0x6f, 0x6e, 0x5f, 0x6d, 0x61, 0x6a, 0x6f, 
	0x72, 0x28, 0x29, 0x0a, 0x20, 0x20, 0x20, 0x20, 0x72, 0x65, 0x74, 
	0x75, 0x72, 0x6e, 0x20, 0x74, 0x6f, 0x6e, 0x75, 0x6d, 0x62, 0x65, 
	0x72, 0x28, 0x6c, 0x69, 0x62, 0x73, 0x2e, 0x73, 0x6f, 0x64, 0x69, 
	0x75, 0x6d, 0x2e, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x5f, 0x6c, 
	0x69, 0x62, 0x72, 0x61, 0x72, 0x79, 0x5f, 0x76, 0x65, 0x72, 0x73, 
	0x69, 0x6f, 0x6e, 0x5f, 0x6d, 0x61, 0x6a, 0x6f, 0x72, 0x28, 0x29, 
	0x29, 0x0a, 0x20, 0x20, 0x65, 0x6e, 0x64, 0x0a, 0x0a, 0x20, 0x20, 
	0x6c, 0x6f, 0x63, 0x61, 0x6c, 0x20, 0x66, 0x75, 0x6e, 0x63, 0x74, 
	0x69, 0x6f, 0x6e, 0x20, 0x6c, 0x73, 0x5f, 0x73, 0x6f, 0x64, 0x69, 
	0x75, 0x6d, 0x5f, 0x6c, 0x69, 0x62, 0x72, 0x61, 0x72, 0x79, 0x5f, 
	0x76, 0x65, 0x72, 0x73, 0x69, 0x6f, 0x6e, 0x5f, 0x6d, 0x69, 0x6e, 
	0x6f, 0x72, 0x28, 0x29, 0x0a, 0x20, 0x20, 0x20, 0x20, 0x72, 0x65, 
	0x74, 0x75, 0x72, 0x6e, 0x20, 0x74, 0x6f, 0x6e, 0x75, 0x6d, 0x62, 
	0x65, 0x72, 0x28, 0x6c, 0x69, 0x62, 0x73, 0x2e, 0x73, 0x6f, 0x64, 
	0x69, 0x75, 0x6d, 0x2e, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x5f, 
	0x6c, 0x69, 0x62, 0x72, 0x61, 0x72, 0x79, 0x5f, 0x76, 0x65, 0x72, 
	0x73, 0x69, 0x6f, 0x6e, 0x5f, 0x6d, 0x69, 0x6e, 0x6f, 0x72, 0x28, 
	0x29, 0x29, 0x0a, 0x20, 0x20, 0x65, 0x6e, 0x64, 0x0a, 0x0a, 0x20, 
	0x20, 0x6c, 0x6f, 0x63, 0x61, 0x6c, 0x20, 0x66, 0x75, 0x6e, 0x63, 
	0x74, 0x69, 0x6f, 0x6e, 0x20, 0x6c, 0x73, 0x5f, 0x73, 0x6f, 0x64, 
	0x69, 0x75, 0x6d, 0x5f, 0x6c, 0x69, 0x62, 0x72, 0x61, 0x72, 0x79, 
	0x5f, 0x6d, 0x69, 0x6e, 0x69, 0x6d, 0x61, 0x6c, 0x28, 0x29, 0x0a, 
	0x20, 0x20, 0x20, 0x20, 0x72, 0x65, 0x74, 0x75, 0x72, 0x6e, 0x20, 
	0x74, 0x6f, 0x6e, 0x75, 0x6d, 0x62, 0x65, 0x72, 0x28, 0x6c, 0x69, 
	0x62, 0x73, 0x2e, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x2e, 0x73, 
	0x6f, 0x64, 0x69, 0x75, 0x6d, 0x5f, 0x6c, 0x69, 0x62, 0x72, 0x61, 
	0x72, 0x79, 0x5f, 0x6d, 0x69, 0x6e, 0x69, 0x6d, 0x61, 0x6c, 0x28, 
	0x29, 0x29, 0x0a, 0x20, 0x20, 0x65, 0x6e, 0x64, 0x0a, 0x0a, 0x20, 
	0x20, 0x4d, 0x2e, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x5f, 0x76, 
	0x65, 0x72, 0x73, 0x69, 0x6f, 0x6e, 0x5f, 0x73, 0x74, 0x72, 0x69, 
	0x6e, 0x67, 0x20, 0x3d, 0x20, 0x6c, 0x73, 0x5f, 0x73, 0x6f, 0x64, 
	0x69, 0x75, 0x6d, 0x5f, 0x76, 0x65, 0x72, 0x73, 0x69, 0x6f, 0x6e, 
	0x5f, 0x73, 0x74, 0x72, 0x69, 0x6e, 0x67, 0x0a, 0x20, 0x20, 0x4d, 
	0x2e, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x5f, 0x6c, 0x69, 0x62, 
	0x72, 0x61, 0x72, 0x79, 0x5f, 0x76, 0x65, 0x72, 0x73, 0x69, 0x6f, 
	0x6e, 0x5f, 0x6d, 0x61, 0x6a, 0x6f, 0x72, 0x20, 0x3d, 0x20, 0x6c, 
	0x73, 0x5f, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x5f, 0x6c, 0x69, 
	0x62, 0x72, 0x61, 0x72, 0x79, 0x5f, 0x76, 0x65, 0x72, 0x73, 0x69, 
	0x6f, 0x6e, 0x5f, 0x6d, 0x61, 0x6a, 0x6f, 0x72, 0x0a, 0x20, 0x20, 
	0x4d, 0x2e, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x5f, 0x6c, 0x69, 
	0x62, 0x72, 0x61, 0x72, 0x79, 0x5f, 0x76, 0x65, 0x72, 0x73, 0x69, 
	0x6f, 0x6e, 0x5f, 0x6d, 0x69, 0x6e, 0x6f, 0x72, 0x20, 0x3d, 0x20, 
	0x6c, 0x73, 0x5f, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x5f, 0x6c, 
	0x69, 0x62, 0x72, 0x61, 0x72, 0x79, 0x5f, 0x76, 0x65, 0x72, 0x73, 
	0x69, 0x6f, 0x6e, 0x5f, 0x6d, 0x69, 0x6e, 0x6f, 0x72, 0x0a, 0x20, 
	0x20, 0x4d, 0x2e, 0x73, 0x6f, 0x64, 0x69, 0x75, 0x6d, 0x5f, 0x6c, 
	0x69, 0x62, 0x72, 0x61, 0x72, 0x79, 0x5f, 0x6d, 0x69, 0x6e, 0x69, 
	0x6d, 0x61, 0x6c, 0x20, 0x3d, 0x20, 0x6c, 0x73, 0x5f, 0x73, 0x6f, 
	0x64, 0x69, 0x75, 0x6d, 0x5f, 0x6c, 0x69, 0x62, 0x72, 0x61, 0x72, 
	0x79, 0x5f, 0x6d, 0x69, 0x6e, 0x69, 0x6d, 0x61, 0x6c, 0x0a, 0x0a, 
	0x20, 0x20, 0x72, 0x65, 0x74, 0x75, 0x72, 0x6e, 0x20, 0x4d, 0x0a, 
	0x65, 0x6e, 0x64, 0x0a, 0x00
};

static const int ls_version_ffi_implementation_length = 1039;

static int
ls_sodium_version_string(lua_State *L) {
    lua_pushstring(L,sodium_version_string());
    return 1;
}

static int
ls_sodium_library_version_major(lua_State *L) {
    lua_pushinteger(L,sodium_library_version_major());
    return 1;
}

static int
ls_sodium_library_version_minor(lua_State *L) {
    lua_pushinteger(L,sodium_library_version_minor());
    return 1;
}

static int
ls_sodium_library_minimal(lua_State *L) {
    lua_pushinteger(L,sodium_library_minimal());
    return 1;
}

static void
ls_version_copydown_table(lua_State *L,int target, int source) {
    lua_pushnil(L);
    while(lua_next(L,source) != 0) {
        lua_setfield(L,target,lua_tostring(L,-2));
    }
}

static int
ls_version_core_setup(lua_State *L) {
    int target = lua_gettop(L);

    if(luaL_loadbuffer(L,ls_version_ffi_implementation,ls_version_ffi_implementation_length - 1, "version.lua")) {
        return lua_error(L);
    }
    if(lua_pcall(L,0,1,0)) {
        return lua_error(L);
    }
    if(lua_pcall(L,0,1,0)) {
        return lua_error(L);
    }

    ls_version_copydown_table(L,target,lua_gettop(L));
    lua_pop(L,1);

    lua_pushcclosure(L,ls_sodium_version_string,0);
    lua_setfield(L,-2,"sodium_version_string");

    lua_pushcclosure(L,ls_sodium_library_version_major,0);
    lua_setfield(L,-2,"sodium_library_version_major");

    lua_pushcclosure(L,ls_sodium_library_version_minor,0);
    lua_setfield(L,-2,"sodium_library_version_minor");

    lua_pushcclosure(L,ls_sodium_library_minimal,0);
    lua_setfield(L,-2,"sodium_library_minimal");

    return 0;
}



int
luaopen_luasodium_core(lua_State *L) {
    LUASODIUM_INIT(L);
    lua_newtable(L);

    ls_crypto_auth_core_setup(L);
    ls_crypto_box_core_setup(L);
    ls_crypto_hash_core_setup(L);
    ls_crypto_onetimeauth_core_setup(L);
    ls_crypto_scalarmult_core_setup(L);
    ls_crypto_secretbox_core_setup(L);
    ls_crypto_sign_core_setup(L);
    ls_crypto_stream_core_setup(L);
    ls_crypto_verify_core_setup(L);
    ls_randombytes_core_setup(L);
    ls_utils_core_setup(L);
    ls_version_core_setup(L);

    return 1;
}
